# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import os
import tempfile
import zipfile
from pathlib import Path

import numpy as np
import pytest
from packaging.version import Version

tifffile = pytest.importorskip("tifffile", reason="tifffile not installed")
hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
t = pytest.importorskip("traits.api", reason="traits not installed")

import rsciio.tiff  # noqa: E402
from rsciio.utils._tests import assert_deep_almost_equal  # noqa: E402

TEST_DATA_PATH = Path(__file__).parent / "data" / "tiff"
TEST_NPZ_DATA_PATH = Path(__file__).parent / "data" / "npz"
TMP_DIR = tempfile.TemporaryDirectory()


def teardown_module():
    TMP_DIR.cleanup()


def _compare_signal_shape_data(s0, s1):
    assert s0.data.shape == s1.data.shape
    np.testing.assert_equal(s0.data, s1.data)


def test_rgba16(tmp_path):
    with zipfile.ZipFile(TEST_DATA_PATH / "test_rgba16.zip", "r") as zipped:
        zipped.extractall(tmp_path)

    s = hs.load(tmp_path / "test_rgba16.tif")
    data = np.load(TEST_NPZ_DATA_PATH / "test_rgba16.npz")["a"]
    assert (s.data == data).all()
    assert s.axes_manager.signal_shape == (128, 128)
    assert s.axes_manager.navigation_shape == (3,)
    assert s.axes_manager[0].units == t.Undefined
    assert s.axes_manager[1].units == t.Undefined
    assert s.axes_manager[2].units == t.Undefined
    np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, atol=1e-5)
    np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, atol=1e-5)
    np.testing.assert_allclose(s.axes_manager[2].scale, 1.0, atol=1e-5)
    assert s.metadata.General.date == "2014-03-31"
    assert s.metadata.General.time == "16:35:46"


class TestDM3ToTiffConversion:
    @staticmethod
    def test_read_unit_um(tmp_path):
        # Load DM file and save it as tif
        s = hs.load(TEST_DATA_PATH / "test_dm_image_um_unit.dm3")
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)
        assert s.metadata.General.date == "2015-07-20"
        assert s.metadata.General.time == "18:48:25"

        fname = tmp_path / "test_export_um_unit.tif"
        s.save(fname, overwrite=True, export_scale=True)
        # load tif file
        s2 = hs.load(fname)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s2.axes_manager[0].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s2.axes_manager[1].scale, 0.16867, atol=1e-5)
        assert s2.metadata.General.date == s.metadata.General.date
        assert s2.metadata.General.time == s.metadata.General.time
        assert s2.axes_manager.signal_shape == s.axes_manager.signal_shape
        assert s2.axes_manager.navigation_shape == s.axes_manager.navigation_shape
        _compare_signal_shape_data(s, s2)

    @staticmethod
    def test_write_read_intensity_axes_DM():
        s = hs.load(TEST_DATA_PATH / "test_dm_image_um_unit.dm3")
        s.metadata.Signal.set_item("quantity", "Electrons (Counts)")
        d = {"gain_factor": 5.0, "gain_offset": 2.0}
        s.metadata.Signal.set_item("Noise_properties.Variance_linear_model", d)
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "tiff_files", "test_export_um_unit2.tif")
            s.save(fname, overwrite=True, export_scale=True)
            s2 = hs.load(fname)
            assert_deep_almost_equal(
                s.metadata.Signal.as_dictionary(), s2.metadata.Signal.as_dictionary()
            )


class TestLoadingImagesSavedWithImageJ:
    @staticmethod
    def test_read_unit_from_imagej():
        s = hs.load(TEST_DATA_PATH / "test_loading_image_saved_with_imageJ.tif")
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)

    @staticmethod
    def test_read_unit_from_imagej_stack():
        s = hs.load(TEST_DATA_PATH / "test_loading_image_saved_with_imageJ_stack.tif")
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == (2,)
        assert s.data.shape == (2, 68, 68)
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == "µm"
        assert s.axes_manager[2].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 2.5, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[2].scale, 0.16867, atol=1e-5)

    @staticmethod
    def test_read_unit_from_imagej_stack_no_scale():
        s = hs.load(
            TEST_DATA_PATH / "test_loading_image_saved_with_imageJ_stack_no_scale.tif"
        )
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == (2,)
        assert s.data.shape == (2, 68, 68)
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == t.Undefined
        assert s.axes_manager[2].units == t.Undefined
        np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[2].scale, 1.0, atol=1e-5)

    @staticmethod
    def test_read_unit_from_imagej_no_scale():
        s = hs.load(
            TEST_DATA_PATH / "test_loading_image_saved_with_imageJ_no_scale.tif"
        )
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == t.Undefined
        np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, atol=1e-5)

    @staticmethod
    def test_write_read_unit_imagej():
        s = hs.load(
            TEST_DATA_PATH / "test_loading_image_saved_with_imageJ.tif",
            convert_units=True,
        )
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].units = "µm"
        with tempfile.TemporaryDirectory() as tmpdir:
            fname2 = os.path.join(tmpdir, "test_loading_image_saved_with_imageJ2.tif")
            s.save(fname2, export_scale=True, overwrite=True)
            s2 = hs.load(fname2)
            assert s2.axes_manager[0].units == "µm"
            assert s2.axes_manager[1].units == "µm"
            assert s.data.shape == s2.data.shape
            assert s2.axes_manager.signal_shape == s.axes_manager.signal_shape
            assert s2.axes_manager.navigation_shape == s.axes_manager.navigation_shape

    @staticmethod
    def test_write_read_unit_imagej_with_description():
        s = hs.load(TEST_DATA_PATH / "test_loading_image_saved_with_imageJ.tif")
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].units = "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)
        with tempfile.TemporaryDirectory() as tmpdir:
            fname2 = os.path.join(tmpdir, "description.tif")
            s.save(fname2, export_scale=False, overwrite=True, description="test")
            s2 = hs.load(fname2)
            assert s2.axes_manager[0].units == t.Undefined
            assert s2.axes_manager[1].units == t.Undefined
            np.testing.assert_allclose(s2.axes_manager[0].scale, 1.0, atol=1e-5)
            np.testing.assert_allclose(s2.axes_manager[1].scale, 1.0, atol=1e-5)
            assert s2.axes_manager.signal_shape == s.axes_manager.signal_shape
            assert s2.axes_manager.navigation_shape == s.axes_manager.navigation_shape

            fname3 = os.path.join(tmpdir, "description2.tif")
            s.save(fname3, export_scale=True, overwrite=True, description="test")
            s3 = hs.load(fname3, convert_units=True)
            assert s3.axes_manager[0].units == "µm"
            assert s3.axes_manager[1].units == "µm"
            np.testing.assert_allclose(s3.axes_manager[0].scale, 0.16867, atol=1e-5)
            np.testing.assert_allclose(s3.axes_manager[1].scale, 0.16867, atol=1e-5)
            assert s3.axes_manager.signal_shape == s.axes_manager.signal_shape
            assert s3.axes_manager.navigation_shape == s.axes_manager.navigation_shape


# Uncomment when there is a proper file handle API
# @pytest.mark.parametrize("size", ((50, 50), (2, 50, 50)))
# def test_lazy_loading(tmp_path, size):
#     dummy_data = np.random.random_sample(size=size)
#     fname = tmp_path / "dummy.tiff"

#     rsciio.tiff.file_writer(fname, {"data": dummy_data})
#     from_tiff = rsciio.tiff.file_reader(fname, lazy=True)
#     data = from_tiff[0]["data"]
#     fh = get_file_handle(data)
#     # check that the file is open
#     fh.fileno()

#     data = data.compute()
#     np.testing.assert_allclose(data, dummy_data)

#     # After we load to memory, we can close the file manually
#     fh.close()
#     with pytest.raises(ValueError):
#         # file is now closed
#         fh.fileno()


# def test_lazy_loading_hyperspy_close(tmp_path):
#     # check that the file is closed automatically in hyperspy
#     dummy_data = np.random.random_sample(size=(2, 50, 50))
#     fname = tmp_path / "dummy.tiff"
#     s = hs.signals.Signal2D(dummy_data)
#     s.save(fname)

#     s2 = hs.load(fname, lazy=True)
#     fh = get_file_handle(s2.data)
#     print("fh", fh)
#     # check that the file is open
#     fh.fileno()
#     s2.compute(close_file=True)
#     np.testing.assert_allclose(s2.data, dummy_data)

#     # when calling compute in hyperspy,
#     # the file should be closed automatically
#     with pytest.raises(ValueError):
#         # file is now closed
#         fh.fileno()


class TestLoadingImagesSavedWithDM:
    @staticmethod
    @pytest.mark.parametrize("lazy", [True, False])
    def test_read_unit_from_DM_stack(tmp_path, lazy):
        s = hs.load(
            TEST_DATA_PATH / "test_loading_image_saved_with_DM_stack.tif", lazy=lazy
        )
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == (2,)
        assert s.data.shape == (2, 68, 68)
        assert s.axes_manager[0].units == "s"
        assert s.axes_manager[1].units == "µm"
        assert s.axes_manager[2].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 2.5, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[2].scale, 1.68674, atol=1e-5)
        fname2 = tmp_path / "test_loading_image_saved_with_DM_stack2.tif"
        s.save(fname2, overwrite=True)
        s2 = hs.load(fname2)
        assert s.axes_manager.signal_shape == s2.axes_manager.signal_shape
        assert s.axes_manager.navigation_shape == s2.axes_manager.navigation_shape
        _compare_signal_shape_data(s, s2)
        assert s2.axes_manager[0].units == s.axes_manager[0].units
        assert s2.axes_manager[1].units == "µm"
        assert s2.axes_manager[2].units == "µm"
        np.testing.assert_allclose(
            s2.axes_manager[0].scale, s.axes_manager[0].scale, atol=1e-5
        )
        np.testing.assert_allclose(
            s2.axes_manager[1].scale, s.axes_manager[1].scale, atol=1e-5
        )
        np.testing.assert_allclose(
            s2.axes_manager[2].scale, s.axes_manager[2].scale, atol=1e-5
        )
        np.testing.assert_allclose(
            s2.axes_manager[0].offset, s.axes_manager[0].offset, atol=1e-5
        )
        np.testing.assert_allclose(
            s2.axes_manager[1].offset, s.axes_manager[1].offset, atol=1e-5
        )
        np.testing.assert_allclose(
            s2.axes_manager[2].offset, s.axes_manager[2].offset, atol=1e-5
        )
        if lazy:
            s.compute()

    @staticmethod
    def test_read_unit_from_dm():
        fname = TEST_DATA_PATH / "test_loading_image_saved_with_DM.tif"
        s = hs.load(fname)
        assert s.axes_manager.signal_shape == (68, 68)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.16867, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[0].offset, 139.66264, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].offset, 128.19276, atol=1e-5)
        with tempfile.TemporaryDirectory() as tmpdir:
            fname2 = os.path.join(tmpdir, "DM2.tif")
            s.save(fname2, overwrite=True)
            s2 = hs.load(fname2)
            _compare_signal_shape_data(s, s2)
            assert s.axes_manager.signal_shape == s2.axes_manager.signal_shape
            assert s.axes_manager.navigation_shape == s2.axes_manager.navigation_shape
            assert s2.axes_manager[0].units == "µm"
            assert s2.axes_manager[1].units == "µm"
            np.testing.assert_allclose(
                s2.axes_manager[0].scale, s.axes_manager[0].scale, atol=1e-5
            )
            np.testing.assert_allclose(
                s2.axes_manager[1].scale, s.axes_manager[1].scale, atol=1e-5
            )
            np.testing.assert_allclose(
                s2.axes_manager[0].offset, s.axes_manager[0].offset, atol=1e-5
            )
            np.testing.assert_allclose(
                s2.axes_manager[1].offset, s.axes_manager[1].offset, atol=1e-5
            )


class TestSavingTiff:
    @staticmethod
    def test_saving_with_custom_tag():
        s = hs.signals.Signal2D(np.arange(10 * 15, dtype=np.uint8).reshape((10, 15)))
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test_saving_with_custom_tag.tif")
            extratag = [(65000, "s", 1, "Random metadata", False)]
            s.save(fname, extratags=extratag, overwrite=True)
            s2 = hs.load(fname)
            assert s2.original_metadata["Number_65000"] == "Random metadata"

    def test_write_scale_unit(self):
        self._test_write_scale_unit(export_scale=True)

    def test_write_scale_unit_no_export_scale(self):
        self._test_write_scale_unit(export_scale=False)

    @staticmethod
    def _test_write_scale_unit(export_scale=True):
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.signals.Signal2D(np.arange(10 * 15, dtype=np.uint8).reshape((10, 15)))
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "y"
        s.axes_manager["x"].scale = 0.25
        s.axes_manager["y"].scale = 0.25
        s.axes_manager["x"].units = "nm"
        s.axes_manager["y"].units = "nm"
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test_export_scale_unit_%s.tif" % export_scale)
            s.save(fname, overwrite=True, export_scale=export_scale)

    @staticmethod
    def test_write_scale_unit_not_square_pixel():
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.signals.Signal2D(np.arange(10 * 15, dtype=np.uint8).reshape((10, 15)))
        s.change_dtype(np.uint8)
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "y"
        s.axes_manager["x"].scale = 0.25
        s.axes_manager["y"].scale = 0.5
        s.axes_manager["x"].units = "nm"
        s.axes_manager["y"].units = "µm"
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test_export_scale_unit_not_square_pixel.tif")
            s.save(fname, overwrite=True, export_scale=True)

    @staticmethod
    def test_write_scale_with_undefined_unit():
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.signals.Signal2D(np.arange(10 * 15, dtype=np.uint8).reshape((10, 15)))
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "y"
        s.axes_manager["x"].scale = 0.25
        s.axes_manager["y"].scale = 0.25
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test_export_scale_undefined_unit.tif")
            s.save(fname, overwrite=True, export_scale=True)

    @staticmethod
    def test_write_scale_with_undefined_scale(tmp_path):
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.signals.Signal2D(np.arange(10 * 15, dtype=np.uint8).reshape((10, 15)))
        fname = tmp_path / "test_export_scale_undefined_scale.tif"
        s.save(fname, overwrite=True, export_scale=True)
        s1 = hs.load(fname)
        _compare_signal_shape_data(s, s1)

    @staticmethod
    def test_write_scale_with_um_unit(tmp_path):
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.load(TEST_DATA_PATH / "test_dm_image_um_unit.dm3")

        fname = tmp_path / "test_export_um_unit.tif"
        s.save(fname, overwrite=True, export_scale=True)
        s1 = hs.load(fname)
        _compare_signal_shape_data(s, s1)

    @staticmethod
    def test_write_scale_unit_image_stack(tmp_path):
        """Lazy test, still need to open the files in ImageJ or DM to check if the
        scale and unit are correct"""
        s = hs.signals.Signal2D(
            np.arange(5 * 10 * 15, dtype=np.uint8).reshape((5, 10, 15))
        )
        s.axes_manager[0].scale = 0.25
        s.axes_manager[1].scale = 0.5
        s.axes_manager[2].scale = 1.5
        s.axes_manager[0].units = "nm"
        s.axes_manager[1].units = "µm"
        s.axes_manager[2].units = "µm"

        fname = tmp_path / "test_export_scale_unit_stack2.tif"
        s.save(fname, overwrite=True, export_scale=True)
        s1 = hs.load(fname, convert_units=True)
        _compare_signal_shape_data(s, s1)
        assert s1.axes_manager[0].units == "pm"
        # only one unit can be read
        assert s1.axes_manager[1].units == "µm"
        assert s1.axes_manager[2].units == "µm"
        np.testing.assert_allclose(s1.axes_manager[0].scale, 250.0)
        np.testing.assert_allclose(s1.axes_manager[1].scale, s.axes_manager[1].scale)
        np.testing.assert_allclose(s1.axes_manager[2].scale, s.axes_manager[2].scale)

    @staticmethod
    def test_saving_loading_stack_no_scale():
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test_export_scale_unit_stack2.tif")
            s0 = hs.signals.Signal2D(np.zeros((10, 20, 30)))
            s0.save(fname, overwrite=True)
            s1 = hs.load(fname)
            _compare_signal_shape_data(s0, s1)


class TestReadFEIHelios:
    path = Path(TMP_DIR.name)

    FEI_Helios_metadata = {
        "Acquisition_instrument": {
            "SEM": {
                "Stage": {
                    "rotation": -2.3611,
                    "tilt": 6.54498e-06,
                    "x": 2.576e-05,
                    "y": -0.000194177,
                    "z": 0.007965,
                },
                "beam_current": 0.00625,
                "beam_energy": 5.0,
                "dwell_time": 1e-05,
                "microscope": 'Helios NanoLab" 660',
                "working_distance": 4.03466,
            }
        },
        "General": {
            "original_filename": "FEI-Helios-Ebeam-8bits.tif",
            "title": "",
            "authors": "supervisor",
            "date": "2016-06-13",
            "time": "17:06:40",
            "FileIO": {
                "0": {
                    "operation": "load",
                    "hyperspy_version": hs.__version__,
                    "io_plugin": "rsciio.tiff",
                }
            },
        },
        "Signal": {"signal_type": ""},
        "_HyperSpy": {
            "Folding": {
                "original_axes_manager": None,
                "original_shape": None,
                "signal_unfolded": False,
                "unfolded": False,
            }
        },
    }

    FEI_navcam_metadata = {
        "_HyperSpy": {
            "Folding": {
                "unfolded": False,
                "signal_unfolded": False,
                "original_shape": None,
                "original_axes_manager": None,
            }
        },
        "General": {
            "original_filename": "FEI-Helios-navcam-with-no-IRBeam.tif",
            "title": "",
            "date": "2022-05-17",
            "time": "09:07:08",
            "authors": "user",
            "FileIO": {"0": {"operation": "load", "io_plugin": "rsciio.tiff"}},
        },
        "Signal": {"signal_type": ""},
        "Acquisition_instrument": {
            "SEM": {
                "Stage": {
                    "x": 0.0699197,
                    "y": 0.000811186,
                    "z": 0,
                    "rotation": 1.07874,
                    "tilt": 6.54498e-06,
                },
                "working_distance": -0.012,
                "microscope": 'Helios NanoLab" 660',
            }
        },
    }

    @classmethod
    def setup_class(cls):
        zipf = TEST_DATA_PATH / "tiff_FEI_Helios.zip"
        with zipfile.ZipFile(zipf, "r") as zipped:
            zipped.extractall(cls.path)

    def test_read_FEI_SEM_scale_metadata_8bits(self):
        fname = self.path / "FEI-Helios-Ebeam-8bits.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (512, 471)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 3.3724, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 3.3724, rtol=1e-5)
        assert s.data.dtype == "uint8"
        # delete timestamp from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        self.FEI_Helios_metadata["General"]["original_filename"] = (
            "FEI-Helios-Ebeam-8bits.tif"
        )
        assert_deep_almost_equal(s.metadata.as_dictionary(), self.FEI_Helios_metadata)

    def test_read_FEI_SEM_scale_metadata_16bits(self):
        fname = self.path / "FEI-Helios-Ebeam-16bits.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (512, 471)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 3.3724, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 3.3724, rtol=1e-5)
        assert s.data.dtype == "uint16"
        # delete timestamp from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        self.FEI_Helios_metadata["General"]["original_filename"] = (
            "FEI-Helios-Ebeam-16bits.tif"
        )
        assert_deep_almost_equal(s.metadata.as_dictionary(), self.FEI_Helios_metadata)

    def test_read_FEI_navcam_metadata(self):
        fname = self.path / "FEI-Helios-navcam.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (768, 551)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "mm"
        assert s.axes_manager[1].units == "mm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.2640, rtol=0.0001)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.2640, rtol=0.0001)
        assert s.data.dtype == "uint8"
        # delete timestamp and version from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        del s.metadata.General.FileIO.Number_0.hyperspy_version
        self.FEI_navcam_metadata["General"]["original_filename"] = (
            "FEI-Helios-navcam.tif"
        )
        assert_deep_almost_equal(s.metadata.as_dictionary(), self.FEI_navcam_metadata)

    def test_read_FEI_navcam_no_IRBeam_metadata(self):
        fname = self.path / "FEI-Helios-navcam-with-no-IRBeam.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (768, 551)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == t.Undefined
        np.testing.assert_allclose(s.axes_manager[0].scale, 1, rtol=0)
        np.testing.assert_allclose(s.axes_manager[1].scale, 1, rtol=0)
        assert s.data.dtype == "uint8"
        # delete timestamp and version from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        del s.metadata.General.FileIO.Number_0.hyperspy_version
        self.FEI_navcam_metadata["General"]["original_filename"] = (
            "FEI-Helios-navcam-with-no-IRBeam.tif"
        )
        assert_deep_almost_equal(s.metadata.as_dictionary(), self.FEI_navcam_metadata)

    def test_read_FEI_navcam_no_IRBeam_bad_floats_metadata(self):
        fname = self.path / "FEI-Helios-navcam-with-no-IRBeam-bad-floats.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (768, 551)
        assert s.axes_manager.navigation_shape == ()
        # delete timestamp and version from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        del s.metadata.General.FileIO.Number_0.hyperspy_version
        self.FEI_navcam_metadata["General"]["original_filename"] = (
            "FEI-Helios-navcam-with-no-IRBeam-bad-floats.tif"
        )

        # working distance in the file was a bogus value,
        # so it shouldn't be in the resulting metadata
        del self.FEI_navcam_metadata["Acquisition_instrument"]["SEM"][
            "working_distance"
        ]
        assert_deep_almost_equal(s.metadata.as_dictionary(), self.FEI_navcam_metadata)


class TestReadZeissSEM:
    path = Path(TMP_DIR.name)

    @classmethod
    def setup_class(cls):
        zipf = TEST_DATA_PATH / "tiff_Zeiss_SEM.zip"
        with zipfile.ZipFile(zipf, "r") as zipped:
            zipped.extractall(cls.path)

    def test_read_Zeiss_SEM_scale_metadata_1k_image(self):
        md = {
            "Acquisition_instrument": {
                "SEM": {
                    "Stage": {
                        "rotation": 10.2,
                        "tilt": -0.0,
                        "x": 75.6442,
                        "y": 60.4901,
                        "z": 25.193,
                    },
                    "Detector": {"detector_type": "HE-SE2"},
                    "beam_current": 1.0,
                    "beam_energy": 25.0,
                    "dwell_time": 5e-08,
                    "magnification": 105.0,
                    "microscope": "Merlin-61-08",
                    "working_distance": 14.808,
                }
            },
            "General": {
                "authors": "LIM",
                "date": "2015-12-23",
                "original_filename": "test_tiff_Zeiss_SEM_1k.tif",
                "time": "09:40:32",
                "title": "",
                "FileIO": {
                    "0": {
                        "operation": "load",
                        "hyperspy_version": hs.__version__,
                        "io_plugin": "rsciio.tiff",
                    }
                },
            },
            "Signal": {"signal_type": ""},
            "_HyperSpy": {
                "Folding": {
                    "original_axes_manager": None,
                    "original_shape": None,
                    "signal_unfolded": False,
                    "unfolded": False,
                }
            },
        }

        fname = self.path / "test_tiff_Zeiss_SEM_1k.tif"
        s = hs.load(fname, convert_units=True)

        assert s.axes_manager.signal_shape == (1024, 768)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 2.614514, rtol=1e-6)
        np.testing.assert_allclose(s.axes_manager[1].scale, 2.614514, rtol=1e-6)
        assert s.data.dtype == "uint8"
        # delete timestamp from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        assert_deep_almost_equal(s.metadata.as_dictionary(), md)

    def test_read_Zeiss_SEM_scale_metadata_512_image(self):
        md = {
            "Acquisition_instrument": {
                "SEM": {
                    "Stage": {
                        "rotation": 245.8,
                        "tilt": 0.0,
                        "x": 62.9961,
                        "y": 65.3168,
                        "z": 44.678,
                    },
                    "beam_energy": 5.0,
                    "magnification": "50.00 K X",
                    "microscope": "ULTRA 55-36-06",
                    "working_distance": 3.9,
                }
            },
            "General": {
                "authors": "LIBERATO",
                "date": "2018-09-25",
                "original_filename": "test_tiff_Zeiss_SEM_512pix.tif",
                "time": "08:20:42",
                "title": "",
                "FileIO": {
                    "0": {
                        "operation": "load",
                        "hyperspy_version": hs.__version__,
                        "io_plugin": "rsciio.tiff",
                    }
                },
            },
            "Signal": {"signal_type": ""},
            "_HyperSpy": {
                "Folding": {
                    "original_axes_manager": None,
                    "original_shape": None,
                    "signal_unfolded": False,
                    "unfolded": False,
                }
            },
        }

        fname = self.path / "test_tiff_Zeiss_SEM_512pix.tif"
        s = hs.load(fname, convert_units=True)
        assert s.axes_manager.signal_shape == (512, 384)
        assert s.axes_manager.navigation_shape == ()
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.011649976, rtol=1e-6)
        np.testing.assert_allclose(s.axes_manager[1].scale, 0.011649976, rtol=1e-6)
        assert s.data.dtype == "uint8"
        # delete timestamp from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        assert_deep_almost_equal(s.metadata.as_dictionary(), md)

    def test_zeiss_multipage_as_separate_signals(self):
        fname = self.path / "ZEISS_SEM_multipage.tif"
        s = hs.load(fname, multipage_as_list=True)
        assert len(s) == 2
        assert s[0].axes_manager.signal_shape == (2048, 1536)
        assert s[0].axes_manager.navigation_shape == ()
        assert s[0].axes_manager.signal_shape == s[1].axes_manager.signal_shape
        assert s[0].axes_manager.navigation_shape == s[1].axes_manager.navigation_shape
        assert s[0].metadata.General.time == "22:42:02"
        assert s[1].metadata.General.time == "22:45:19"
        assert s[0].original_metadata.CZ_SEM.ap_fib_slice_index == ("Slice Index", 2)
        assert s[1].original_metadata.CZ_SEM.ap_fib_slice_index == ("Slice Index", 5)
        # ('C3 Lens I', 716.29, 'mA')
        # ('C3 Lens I', 716.28, 'mA')
        np.testing.assert_allclose(
            s[0].original_metadata.CZ_SEM.ap_c3[1], 716.29, rtol=1e-6
        )
        np.testing.assert_allclose(
            s[1].original_metadata.CZ_SEM.ap_c3[1], 716.28, rtol=1e-6
        )
        np.testing.assert_allclose(
            s[0].metadata.Acquisition_instrument.SEM.working_distance, 4.5, rtol=1e-3
        )
        np.testing.assert_allclose(
            s[1].metadata.Acquisition_instrument.SEM.working_distance, 4.5, rtol=1e-3
        )
        # yes, working distance is such a low resolution record of beam focus...


class TestReadZeissAxioVision:
    @staticmethod
    def test_read_RGB_Zeiss_optical_scale_metadata():
        s = hs.load(TEST_DATA_PATH / "optical_Zeiss_AxioVision_RGB.tif")
        assert s.axes_manager.signal_shape == (13, 10)
        assert s.axes_manager.navigation_shape == ()
        dtype = np.dtype([("R", "u1"), ("G", "u1"), ("B", "u1")])
        assert s.data.dtype == dtype
        assert s.data.shape == (10, 13)
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == t.Undefined
        np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, rtol=1e-5)
        assert s.metadata.General.date == "2016-06-13"
        assert s.metadata.General.time == "15:59:52"
        assert s.metadata.General.FileIO.Number_0.hyperspy_version == hs.__version__
        assert s.metadata.General.FileIO.Number_0.io_plugin == "rsciio.tiff"

    @staticmethod
    def test_read_BW_Zeiss_optical_scale_metadata():
        s = hs.load(
            TEST_DATA_PATH / "optical_Zeiss_AxioVision_BW.tif",
            force_read_resolution=True,
            convert_units=True,
        )
        assert s.axes_manager.signal_shape == (13, 10)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.dtype == np.uint8
        assert s.data.shape == (10, 13)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 169.333, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 169.333, rtol=1e-5)
        assert s.metadata.General.date == "2016-06-13"
        assert s.metadata.General.time == "16:08:49"

    @staticmethod
    def test_read_BW_Zeiss_optical_scale_metadata_convert_units_false():
        s = hs.load(
            TEST_DATA_PATH / "optical_Zeiss_AxioVision_BW.tif",
            force_read_resolution=True,
            convert_units=False,
        )
        assert s.axes_manager.signal_shape == (13, 10)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.dtype == np.uint8
        assert s.data.shape == (10, 13)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 169.333, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 169.333, rtol=1e-5)

    @staticmethod
    def test_read_BW_Zeiss_optical_scale_metadata2():
        s = hs.load(
            TEST_DATA_PATH / "optical_Zeiss_AxioVision_BW.tif",
            force_read_resolution=True,
            convert_units=True,
        )
        assert s.axes_manager.signal_shape == (13, 10)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.dtype == np.uint8
        assert s.data.shape == (10, 13)
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"
        np.testing.assert_allclose(s.axes_manager[0].scale, 169.333, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 169.333, rtol=1e-5)
        assert s.metadata.General.date == "2016-06-13"
        assert s.metadata.General.time == "16:08:49"

    @staticmethod
    def test_read_BW_Zeiss_optical_scale_metadata3():
        s = hs.load(
            TEST_DATA_PATH / "optical_Zeiss_AxioVision_BW.tif",
            force_read_resolution=False,
        )
        assert s.axes_manager.signal_shape == (13, 10)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.dtype == np.uint8
        assert s.data.shape == (10, 13)
        assert s.axes_manager[0].units == t.Undefined
        assert s.axes_manager[1].units == t.Undefined
        np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, rtol=1e-5)
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, rtol=1e-5)
        assert s.metadata.General.date == "2016-06-13"
        assert s.metadata.General.time == "16:08:49"


def test_read_TVIPS_metadata(tmp_path):
    md = {
        "Acquisition_instrument": {
            "TEM": {
                "Detector": {"Camera": {"exposure": 0.4, "name": "F416"}},
                "Stage": {
                    "tilt_alpha": -0.0070000002,
                    "tilt_beta": -0.055,
                    "x": 0.0,
                    "y": -9.2000000506686774e-05,
                    "z": 7.0000001350933871e-06,
                },
                "beam_energy": 99.0,
                "magnification": 32000.0,
            }
        },
        "General": {
            "original_filename": "TVIPS_bin4.tif",
            "time": "9:01:17",
            "title": "",
            "FileIO": {
                "0": {
                    "operation": "load",
                    "hyperspy_version": hs.__version__,
                    "io_plugin": "rsciio.tiff",
                }
            },
        },
        "Signal": {"signal_type": ""},
        "_HyperSpy": {
            "Folding": {
                "original_axes_manager": None,
                "original_shape": None,
                "signal_unfolded": False,
                "unfolded": False,
            }
        },
    }

    zipf = TEST_DATA_PATH / "TVIPS_bin4.zip"
    with zipfile.ZipFile(zipf, "r") as zipped:
        zipped.extractall(tmp_path)
        s = hs.load(tmp_path / "TVIPS_bin4.tif", convert_units=True)
    assert s.axes_manager.signal_shape == (1024, 1024)
    assert s.axes_manager.navigation_shape == ()
    assert s.data.dtype == np.uint8
    assert s.data.shape == (1024, 1024)
    assert s.axes_manager[0].units == "nm"
    assert s.axes_manager[1].units == "nm"
    np.testing.assert_allclose(s.axes_manager[0].scale, 1.42080, rtol=1e-5)
    np.testing.assert_allclose(s.axes_manager[1].scale, 1.42080, rtol=1e-5)
    # delete timestamp from metadata since it's runtime dependent
    del s.metadata.General.FileIO.Number_0.timestamp
    assert_deep_almost_equal(s.metadata.as_dictionary(), md)


def test_axes_metadata():
    data = np.arange(2 * 5 * 10).reshape((2, 5, 10))
    s = hs.signals.Signal2D(data)
    nav_unit = "s"
    s.axes_manager.navigation_axes[0].units = nav_unit
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, "axes_metadata_default.tif")
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.axes_manager.navigation_axes[0].name == "image series"
        assert s2.axes_manager.navigation_axes[0].units == nav_unit
        assert s2.axes_manager.navigation_axes[0].is_binned is False

        fname2 = os.path.join(tmpdir, "axes_metadata_IYX.tif")
        s.save(fname2, metadata={"axes": "IYX"})
        s3 = hs.load(fname2)
        assert s3.axes_manager.navigation_axes[0].name == "image series"
        assert s3.axes_manager.navigation_axes[0].units == nav_unit
        assert s3.axes_manager.navigation_axes[0].is_binned is False

        fname2 = os.path.join(tmpdir, "axes_metadata_ZYX.tif")
        s.save(fname2, metadata={"axes": "ZYX"})
        s3 = hs.load(fname2)
        assert s3.axes_manager.navigation_axes[0].units == nav_unit
        assert s3.axes_manager.navigation_axes[0].is_binned is False


def test_olympus_SIS():
    pytest.importorskip("imagecodecs", reason="imagecodecs is required")
    fname = TEST_DATA_PATH / "olympus_SIS.tif"
    s = hs.load(fname)
    # This olympus SIS contains two images:
    # - the first one is a RGB 8-bits (used for preview purposes)
    # - the second one is the raw data
    # only the second one is calibrated.
    assert len(s) == 2
    assert s[0].axes_manager.signal_shape == (112, 101)
    assert s[0].axes_manager.navigation_shape == ()
    assert s[0].axes_manager.signal_shape == s[1].axes_manager.signal_shape
    assert s[0].axes_manager.navigation_shape == s[1].axes_manager.navigation_shape
    am = s[1].axes_manager
    for axis in am._axes:
        assert axis.units == "m"
        np.testing.assert_allclose(axis.scale, 2.3928e-11)
        np.testing.assert_allclose(axis.offset, 0.0)

    for ima in s:
        assert ima.data.shape == (101, 112)

    assert s[1].data.dtype is np.dtype("uint16")


def test_save_angstrom_units():
    s = hs.signals.Signal2D(np.arange(200 * 200, dtype="float32").reshape((200, 200)))
    for axis in s.axes_manager.signal_axes:
        axis.units = "Å"
        axis.scale = 0.1
        axis.offset = 10

    with tempfile.TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, "save_angstrom_units.tif")
        s.save(fname)
        s2 = hs.load(fname)
        if Version(tifffile.__version__) >= Version("2020.7.17"):
            assert s2.axes_manager[0].units == s.axes_manager[0].units
        assert s2.axes_manager[0].scale == s.axes_manager[0].scale
        assert s2.axes_manager[0].offset == s.axes_manager[0].offset
        assert s2.axes_manager[0].is_binned == s.axes_manager[0].is_binned


def test_JEOL_SightX(tmp_path):
    files = [
        ("JEOL-SightX-Ronchigram-dummy.tif.gz", 1.0, t.Undefined),
        ("JEOL-SightX-SAED-dummy.tif.gz", 0.2723, "1 / nm"),
        ("JEOL-SightX-TEM-mag-dummy.tif.gz", 1.8208, "nm"),
    ]
    for file in files:
        fname = file[0]
        if fname[-3:] == ".gz":
            import gzip

            with gzip.open(TEST_DATA_PATH / fname, "rb") as f:
                content = f.read()
            fname = tmp_path / fname[:-3]
            with open(fname, "wb") as f2:
                f2.write(content)
            s = hs.load(fname)
        else:
            s = hs.load(TEST_DATA_PATH / file[0])
        assert s.axes_manager.signal_shape == (556, 556)
        assert s.axes_manager.navigation_shape == ()
        for i in range(2):  # x, y
            assert s.axes_manager[i].size == 556
            np.testing.assert_allclose(s.axes_manager[i].scale, file[1], rtol=1e-3)
            assert s.axes_manager[i].units == file[2]


class TestReadHamamatsu:
    path = Path(TMP_DIR.name)

    @classmethod
    def setup_class(cls):
        zipf = TEST_DATA_PATH / "tiff_hamamatsu.zip"
        with zipfile.ZipFile(zipf, "r") as zipped:
            zipped.extractall(cls.path)

    def test_hamamatsu_streak_loadwarnings(self):
        file = "test_hamamatsu_streak_SCAN.tif"
        fname = os.path.join(self.path, file)

        # No hamamatsu_streak_axis_type argument should :
        # - raise warning
        # - Initialise uniform data axis
        with pytest.warns(UserWarning):
            s = hs.load(fname)
            assert s.axes_manager.all_uniform

        # Invalid hamamatsu_streak_axis_type argument should:
        # - raise warning
        # - Initialise uniform data axis
        with pytest.raises(ValueError):
            _ = hs.load(fname, hamamatsu_streak_axis_type="xxx")

        s = hs.load(fname, hamamatsu_streak_axis_type="uniform")
        assert s.axes_manager.all_uniform

    def test_hamamatsu_streak_scanfile(self):
        file = "test_hamamatsu_streak_SCAN.tif"
        fname = os.path.join(self.path, file)

        with pytest.warns(UserWarning):
            s = hs.load(fname)

        assert s.axes_manager.signal_shape == (672, 508)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.shape == (508, 672)
        assert s.axes_manager[1].units == "ps"
        np.testing.assert_allclose(s.axes_manager[1].scale, 2.3081, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[1].offset, 652.3756, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.01714, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[0].offset, 231.0909, rtol=1e-3)

    def test_hamamatsu_streak_focusfile(self):
        file = "test_hamamatsu_streak_FOCUS.tif"
        fname = os.path.join(self.path, file)

        with pytest.warns(UserWarning):
            s = hs.load(fname)

        assert s.axes_manager.signal_shape == (672, 508)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.shape == (508, 672)
        assert s.axes_manager[1].units == ""
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[1].offset, 0.0, rtol=1e-3, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[0].scale, 0.01714, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[0].offset, 231.0909, rtol=1e-3)

    def test_hamamatsu_streak_otherfile(self):
        # Test case where ScalingXScalingFile="Other"
        file = "test_hamamatsu_streak_OTHER.tif"
        fname = os.path.join(self.path, file)

        with pytest.warns(UserWarning):
            s = hs.load(fname)

        assert s.axes_manager.signal_shape == (672, 508)
        assert s.axes_manager.navigation_shape == ()
        assert s.data.shape == (508, 672)
        assert s.axes_manager[0].units == ""
        assert s.axes_manager[1].units == ""
        np.testing.assert_allclose(s.axes_manager[1].scale, 1.0, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[1].offset, 0.0, rtol=1e-3, atol=1e-5)
        np.testing.assert_allclose(s.axes_manager[0].scale, 1.0, rtol=1e-3)
        np.testing.assert_allclose(s.axes_manager[0].offset, 0.0, rtol=1e-3, atol=1e-5)

    def test_hamamatsu_streak_non_uniform_load(self):
        file = "test_hamamatsu_streak_SCAN.tif"
        fname = os.path.join(self.path, file)

        s = hs.load(fname, hamamatsu_streak_axis_type="data")

        np.testing.assert_allclose(
            s.original_metadata.ImageDescriptionParsed.Scaling.ScalingYaxis,
            s.axes_manager[1].axis,
            rtol=1e-5,
        )

        s = hs.load(fname, hamamatsu_streak_axis_type="functional")

        np.testing.assert_allclose(
            s.original_metadata.ImageDescriptionParsed.Scaling.ScalingYaxis,
            s.axes_manager[1].axis,
            rtol=1e-5,
        )

    def test_is_hamamatsu_streak(self):
        file = "test_hamamatsu_streak_SCAN.tif"
        fname = os.path.join(self.path, file)

        with pytest.warns(UserWarning):
            s = hs.load(fname)

        omd = s.original_metadata.as_dictionary()

        omd["Artist"] = "TAPTAP"

        assert not rsciio.tiff._api._is_streak_hamamatsu(omd)

        _ = omd.pop("Artist")

        assert not rsciio.tiff._api._is_streak_hamamatsu(omd)

        omd.update({"Artist": "Copyright Hamamatsu GmbH, 2018"})

        omd["Software"] = "TAPTAPTAP"

        assert not rsciio.tiff._api._is_streak_hamamatsu(omd)

        _ = omd.pop("Software")

        assert not rsciio.tiff._api._is_streak_hamamatsu(omd)

        omd.update({"Software": "HPD-TA 9.5 pf4"})

        assert rsciio.tiff._api._is_streak_hamamatsu(omd)

    def test_lumispy(self):
        pytest.importorskip("lumispy", reason="lumispy not installed")
        file = "test_hamamatsu_streak_SCAN.tif"
        fname = os.path.join(self.path, file)

        with pytest.warns(UserWarning):
            s = hs.load(fname)

        assert s.metadata.Signal.signal_type == "TransientSpectrum"



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib
import logging
import os
from collections.abc import Iterable
from datetime import datetime, timedelta

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC, RETURNS_DOC
from rsciio.utils._decorator import jit_ifnumba

_logger = logging.getLogger(__name__)


jTYPE = {
    1: "B",
    2: "H",
    3: "i",
    4: "f",
    5: "d",
    6: "B",
    7: "H",
    8: "i",
    9: "f",
    10: "d",
    11: "?",
    12: "c",
    13: "c",
    14: "H",
    20: "c",
    65553: "?",
    65552: "?",
}


def file_reader(
    filename,
    lazy=False,
    rebin_energy=1,
    sum_frames=True,
    SI_dtype=np.uint8,
    cutoff_at_kV=None,
    downsample=1,
    only_valid_data=True,
    read_em_image=False,
    frame_list=None,
    frame_shifts=None,
    frame_start_index=None,
):
    """
    File reader for JEOL Analysist Station software format.

    Parameters
    ----------
    %s
    %s
    rebin_energy : int, default=1
        Factor used to rebin the energy dimension. It must be a
        divisor of the number of channels, typically 4096.
    sum_frames : bool, default=True
        If ``False``, each individual frame (sweep in JEOL software jargon)
        is loaded. Be aware that loading each individual frame will use a lot of memory.
        However, it can be used in combination with ``rebin_energy``, ``cutoff_at_kV``
        and ``downsample`` to reduce memory usage.
    SI_dtype : numpy.dtype, default=np.uint8
        Set the dtype of the eds dataset. Useful to adjust memory usage
        and maximum number of X-rays per channel.
    cutoff_at_kV : int, float, None, default=None
        If set (>= 0), use to crop the energy range up to the specified energy.
        If ``None``, the whole energy range is loaded.
        Useful to reduce memory usage.
    downsample : int, default=1
        The downsample ratio of the navigation dimension of an EDS
        dataset. It can be an integer or a tuple of length 2 to define ``x`` and ``y``
        separately and it must be a divisor of the size of the navigation dimension.
    only_valid_data : bool, default=True
        For ``.pts`` files only. Ignore incomplete and partly
        acquired last frame, which typically occurs when the acquisition was
        interrupted. When loading incomplete data (``only_valid_data=False``),
        the missing data are filled with zeros. If ``sum_frames=True``, this argument
        will be ignored to enforce consistent summing over the mapped area.
    read_em_image : bool, default=False
        For ``.pts`` files only. If ``True``,
        read SEM/STEM image from ``.pts`` file if available. In this case, both
        the spectrum Image and SEM/STEM Image will be returned as list.
    frame_list : list of int, None, default=None
        For ``.pts`` files only. Frames in ``frame_list`` will be loaded.
        For example, ``frame_list=[1,3]`` means second and forth frame will be loaded.
        If ``None``, all frames are loaded.
    frame_shifts : list of [int, int], list of [int, int, int], None, default=None
        For ``.pts`` files only. Each frame will be loaded with offset of
        [dy, dx (, and optionary dEnergy)]. Units are pixels/channels.
        The result of estimate_shift2D() can be used as a parameter of frame_shifts.
        This is useful for express drift correction. Not suitable for accurate analysis.
    frame_start_index : list, None, default=None
        The list of offset pointers of each frame in the raw data.
        The pointer for frame0 is 0.

    %s
    """
    kwargs = dict(
        lazy=lazy,
        rebin_energy=rebin_energy,
        sum_frames=sum_frames,
        SI_dtype=SI_dtype,
        cutoff_at_kV=cutoff_at_kV,
        downsample=downsample,
        only_valid_data=only_valid_data,
        read_em_image=read_em_image,
        frame_list=frame_list,
        frame_shifts=frame_shifts,
        frame_start_index=frame_start_index,
    )
    file_ext = os.path.splitext(filename)[-1][1:].lower()
    if file_ext in extension_to_reader_mapping:
        return extension_to_reader_mapping[file_ext](filename, **kwargs)
    else:
        _logger.info(f"{filename} : File type {file_ext} is not supported. Skipping")
        return []


def _read_asw(filename, **kwargs):
    image_list = []
    with open(filename, "br") as fd:
        file_magic = np.fromfile(fd, "<I", 1)[0]
        if file_magic != 0:
            raise ValueError(f"Not a valid JEOL asw format '{filename}'")
        fd.seek(12)
        filetree = _parsejeol(fd)

    filepath, filen = os.path.split(os.path.abspath(filename))
    if "SampleInfo" in filetree:
        for i in filetree["SampleInfo"].keys():
            if "ViewInfo" in filetree["SampleInfo"][i]:
                for j in filetree["SampleInfo"][i]["ViewInfo"].keys():
                    node = filetree["SampleInfo"][i]["ViewInfo"][j]
                    if "ViewData" in node:
                        for k in node["ViewData"].keys():
                            node2 = node["ViewData"][k]
                            # path tuple contains:
                            # root, sample_folder, view_folder, data_file
                            path = node["ViewData"][k]["Filename"].split("\\")
                            subfile = os.path.join(*path)
                            file_path = os.path.join(filepath, subfile)
                            dict_list = file_reader(file_path, **kwargs)
                            for d in dict_list:
                                d["original_metadata"]["asw"] = filetree
                                d["original_metadata"]["asw_viewdata"] = node2
                                image_list.append(d)
                    else:
                        _logger.warning(
                            f"{filename} : SampleInfo[{i}].ViewInfo[{j}] does not have ViewData section."
                        )
            else:
                _logger.warning(
                    f"{filename} : SampleInfo[{i}] does not have ViewInfo section."
                )
    else:
        _logger.warning(f"{filename} does not have SampleInfo section.")

    return image_list


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def _read_img(filename, **kwargs):
    """
    Parameters
    ----------
    filename : str
        img file name
    kwargs : dict
        Not used.

    Returns
    -------
    image_list : list of images(dict)
        (Always len(image_list) == 1)
    """

    with open(filename, "br") as fd:
        file_magic = np.fromfile(fd, "<I", 1)[0]
        if file_magic != 52:
            _logger.warning(f"Not a valid JEOL img format '{filename}'")
            return []

        # fileformat
        _ = _decode(fd.read(32).rstrip(b"\x00"))
        head_pos, head_len, data_pos = np.fromfile(fd, "<I", 3)
        fd.seek(data_pos + 12)
        header_long = _parsejeol(fd)

    width, height = header_long["Image"]["Size"]
    header_long["Image"]["Bits"].resize((height, width))
    data = header_long["Image"]["Bits"]

    scale = (
        header_long["Instrument"]["ScanSize"] / header_long["Instrument"]["Mag"] * 1.0e3
    )
    xscale = scale / width
    yscale = scale / height
    units = "µm"

    axes = [
        {
            "name": "y",
            "size": height,
            "offset": 0,
            "scale": yscale,
            "units": units,
            "navigate": False,
        },
        {
            "name": "x",
            "size": width,
            "offset": 0,
            "scale": xscale,
            "units": units,
            "navigate": False,
        },
    ]

    datefile = datetime(1899, 12, 30) + timedelta(days=header_long["Image"]["Created"])
    hv = header_long["Instrument"]["AccV"]
    if hv <= 30.0:
        mode = "SEM"
    else:
        mode = "TEM"

    metadata = {
        "Acquisition_instrument": {
            mode: {
                "beam_energy": hv,
                "magnification": header_long["Instrument"]["Mag"],
            },
        },
        "General": {
            "original_filename": os.path.basename(filename),
            "date": datefile.date().isoformat(),
            "time": datefile.time().isoformat(),
            "title": header_long["Image"]["Title"],
        },
        "Signal": {},
    }

    dictionary = {
        "data": data,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": header_long,
    }
    return [dictionary]


def _read_pts(
    filename,
    rebin_energy=1,
    sum_frames=True,
    SI_dtype=np.uint8,
    cutoff_at_kV=None,
    downsample=1,
    only_valid_data=True,
    read_em_image=False,
    frame_list=None,
    frame_start_index=None,
    frame_shifts=None,
    lazy=False,
    **kwargs,
):
    """
    Parameters
    ----------
    filename : str
        pts file name
    rebin_energy : int
        Binning parameter along energy axis. Must be 2^n.
    sum_frames : bool
        If False, returns each frame.
    SI_dtype : numpy.dtype
        The data type for spectrum image. Default is numpy.uint8
    cutoff_at_kV : float
        The maximum energy. Useful to reduce memory size of spectrum image.
        Default is None (no cutoff)
    downsample : int or (int, int)
        Downsample along spatial axes to reduce memory size of spectrum image.
        Value must be 2^n. Default is 1 (no downsampling).
    only_valid_data : bool, default True
        Read incomplete frame if only_valid_data == False
        (usually interrupting measurement makes incomplete frame)
    read_em_image : bool, default False
        Read SEM/STEM image from pts file if read_em_image == True
    frame_list : list of int, default None
        List of frame numbers to be read (None for all data)
    frame_shifts : list of [int, int] or list of [int, int, int], default None
        Each frame will be loaded with offset of dy, dx, (and optional energy
        axis). Units are pixels/channels.
        This is useful for express drift correction. Not suitable for accurate
        analysis.
        Like the result of estimate_shift2D(), the first parameter is for y-axis
    frame_start_index: list
        The list of offset pointers of each frame in the raw data.
        The pointer for frame0 is 0.
    lazy : bool, default False
        Read spectrum image into sparse array if lazy == True
        SEM/STEM image is always read into dense array (numpy.ndarray)
    **kwargs : dict
        Not used.

    Returns
    -------
    image_list : list of images(dict)
        The dictionary used to create the signals, list of dictionaries of
        spectrum image and SEM/STEM image if read_em_image == True
    """
    # Needs numba to read pts file
    # We check at this stage to allow reading other JEOL file without numba
    # it could be possible to read pts without numba (non-lazy data only)
    # but there are limited benefit and it keep "things more simple"
    importlib.import_module("numba")

    def _check_divisor(factor, number, string):
        if factor > 1 and number % factor != 0:
            raise ValueError(f"`{string}`({factor}) must be a divisor of {number}.")

    _check_divisor(rebin_energy, 4096, "rebin_energy")
    rebin_energy = int(rebin_energy)

    with open(filename, "br") as fd:
        file_magic = np.fromfile(fd, "<I", 1)[0]
        if file_magic != 304:
            _logger.warning(f"Not a valid JEOL pts format '{filename}'")
            return []

        # fileformat
        _ = _decode(fd.read(8).rstrip(b"\x00"))
        a, b, head_pos, head_len, data_pos, data_len = np.fromfile(fd, "<I", 6)
        # groupname
        _ = _decode(fd.read(128).rstrip(b"\x00"))
        # memo
        _ = _decode(fd.read(132).rstrip(b"\x00"))
        datefile = datetime(1899, 12, 30) + timedelta(days=np.fromfile(fd, "d", 1)[0])
        fd.seek(head_pos + 12)
        header = _parsejeol(fd)
        meas_data_header = header["PTTD Data"]["AnalyzableMap MeasData"]
        width, height = meas_data_header["Meas Cond"]["Pixels"].split("x")
        width = int(width)
        height = int(height)

        if isinstance(downsample, Iterable):
            if len(downsample) > 2:
                raise ValueError(
                    "`downsample` can't be an iterable of length different from 2."
                )
            downsample_width = downsample[0]
            downsample_height = downsample[1]
            _check_divisor(downsample_width, width, "downsample[0]")
            _check_divisor(downsample_height, height, "downsample[1]")
        else:
            downsample_width = downsample_height = int(downsample)
            _check_divisor(downsample_width, width, "downsample")
            _check_divisor(downsample_height, height, "downsample")

        _check_divisor(downsample_width, width, "downsample[0]")
        _check_divisor(downsample_height, height, "downsample[1]")

        # Normalisation factor for the x and y position in the stream; depends
        # on the downsampling and the size of the navigation space
        width_norm = int(4096 / width * downsample_width)
        height_norm = int(4096 / height * downsample_height)

        width = int(width / downsample_width)
        height = int(height / downsample_height)

        channel_number = int(4096 / rebin_energy)

        fd.seek(data_pos)
        # read spectrum image
        rawdata = np.fromfile(fd, dtype="u2")

        scale = (
            header["PTTD Param"]["Params"]["PARAMPAGE0_SEM"]["ScanSize"]
            / meas_data_header["MeasCond"]["Mag"]
            * 1.0e3
        )
        xscale = scale / width
        yscale = scale / height
        units = "µm"

        ch_mod = meas_data_header["Meas Cond"]["Tpl"]
        ch_res = meas_data_header["Doc"]["CoefA"]
        ch_ini = meas_data_header["Doc"]["CoefB"]
        ch_pos = header["PTTD Param"]["Params"]["PARAMPAGE1_EDXRF"]["Tpl"][ch_mod][
            "DigZ"
        ]

        energy_offset = ch_ini - ch_res * ch_pos
        energy_scale = ch_res * rebin_energy

        if cutoff_at_kV is not None:
            channel_number = int(
                np.round((cutoff_at_kV - energy_offset) / energy_scale)
            )
        # pixel time in milliseconds
        pixel_time = meas_data_header["Doc"]["DwellTime(msec)"]

        # Sweep value is not reliable, so +1 frame is needed if sum_frames = False
        # priority of the length of frame_start_index is higher than "sweep" in header
        sweep = meas_data_header["Doc"]["Sweep"]

        auto_frame_list = False
        if frame_list:
            frame_list = np.asarray(frame_list)
        else:
            auto_frame_list = True
            frame_list = np.arange(sweep + 1)

        # Remove frame numbers outside the data range.
        # The frame with number == sweep is accepted in this stage
        #    for incomplete frame
        # If "frame_shifts" option is used, frame number must be in range of frame_shifts
        if frame_shifts is not None:
            nsf = len(frame_shifts)
            wrong_frames_list = frame_list[
                np.where(
                    (frame_list < 0)
                    | (frame_list > sweep)
                    | (frame_list > nsf)
                    | ((frame_list == nsf) & (not auto_frame_list))
                )
            ]
            frame_list = frame_list[
                np.where((0 <= frame_list) & (frame_list <= sweep) & (frame_list < nsf))
            ]
        else:
            wrong_frames_list = frame_list[
                np.where((frame_list < 0) | (frame_list > sweep))
            ]
            frame_list = frame_list[np.where((0 <= frame_list) & (frame_list <= sweep))]

        if len(wrong_frames_list) > 0:
            wrong_frames = wrong_frames_list.flatten().tolist()
            _logger.warning(
                f"Invalid frame number is specified. The frame {wrong_frames} is not found in pts data."
            )

        # + 1 for incomplete frame
        max_frame = frame_list.max() + 1

        if frame_start_index is None:
            frame_start_index = np.full(max_frame, -1, dtype=np.int32)
            frame_start_index[0] = 0
        else:
            frame_start_index = np.asarray(frame_start_index)

        # fill with -1 as invalid index (not loaded)
        if frame_start_index.size < max_frame:
            fi = np.full(max_frame, -1, dtype=np.int32)
            fi[0 : frame_start_index.size] = frame_start_index
            frame_start_index = fi

        if frame_shifts is None:
            frame_shifts = np.zeros((max_frame, 3), dtype=np.int16)

        if len(frame_shifts[0]) == 2:  # fill z with 0
            fr = np.zeros((max_frame, 3), dtype=np.int16)
            fr[: len(frame_shifts), 0:2] = np.asarray(frame_shifts)
            frame_shifts = fr

        (
            data,
            em_data,
            has_em_image,
            sweep,
            frame_start_index,
            last_valid,
            origin,
            frame_shifts_1,
        ) = _readcube(
            rawdata,
            frame_start_index,
            frame_list,
            width,
            height,
            channel_number,
            width_norm,
            height_norm,
            rebin_energy,
            SI_dtype,
            sweep,
            frame_shifts,
            sum_frames,
            read_em_image,
            only_valid_data,
            lazy,
        )
        header["jeol_pts_frame_origin"] = origin
        header["jeol_pts_frame_shifts"] = frame_shifts_1
        header["jeol_pts_frame_start_index"] = frame_start_index
        # axes_em for SEM/STEM image  intensity[(frame,) y, x]
        # axes for spectrum image  count[(frame,) y, x, energy]
        if sum_frames:
            axes_em = []
            width = data.shape[1]
            height = data.shape[0]
        else:
            axes_em = [
                {
                    "index_in_array": 0,
                    "name": "Frame",
                    "size": sweep,
                    "offset": 0,
                    "scale": pixel_time * height * width / 1e3,
                    "units": "s",
                    "navigate": True,
                }
            ]
            width = data.shape[2]
            height = data.shape[1]
        axes_em.extend(
            [
                {
                    "name": "y",
                    "size": height,
                    "offset": origin[1],
                    "scale": yscale,
                    "units": units,
                    "navigate": False,
                },
                {
                    "name": "x",
                    "size": width,
                    "offset": origin[0],
                    "scale": xscale,
                    "units": units,
                    "navigate": False,
                },
            ]
        )
        axes = axes_em.copy()
        for ax in axes:
            ax["navigate"] = True
        axes.append(
            {
                "name": "Energy",
                "size": channel_number,
                "offset": energy_offset,
                "scale": energy_scale,
                "units": "keV",
                "navigate": False,
            },
        )
        if (not last_valid) and only_valid_data:
            _logger.info(
                "The last frame (sweep) is incomplete because the acquisition stopped during this frame. The partially acquired frame is ignored. Use 'sum_frames=False, only_valid_data=False' to read all frames individually, including the last partially completed frame."
            )

        hv = meas_data_header["MeasCond"]["AccKV"]
        if hv <= 30.0:
            mode = "SEM"
        else:
            mode = "TEM"

        detector_hearder = header["PTTD Param"]["Params"]["PARAMPAGE0_SEM"]
        metadata = {
            "Acquisition_instrument": {
                mode: {
                    "beam_energy": hv,
                    "magnification": meas_data_header["MeasCond"]["Mag"],
                    "Detector": {
                        "EDS": {
                            "azimuth_angle": detector_hearder["DirAng"],
                            "detector_type": detector_hearder["DetT"],
                            "elevation_angle": detector_hearder["ElevAng"],
                            "energy_resolution_MnKa": detector_hearder["MnKaRES"],
                            "real_time": meas_data_header["Doc"]["RealTime"],
                        },
                    },
                },
            },
            "General": {
                "original_filename": os.path.basename(filename),
                "date": datefile.date().isoformat(),
                "time": datefile.time().isoformat(),
                "title": "EDS extracted from " + os.path.basename(filename),
            },
            "Signal": {
                "quantity": "X-rays (Counts)",
                "signal_type": "EDS_" + mode,
            },
        }
        metadata_em = {
            "Acquisition_instrument": {
                mode: {
                    "beam_energy": hv,
                    "magnification": meas_data_header["MeasCond"]["Mag"],
                },
            },
            "General": {
                "original_filename": os.path.basename(filename),
                "date": datefile.date().isoformat(),
                "time": datefile.time().isoformat(),
                "title": "S(T)EM Image extracted from " + os.path.basename(filename),
            },
            "Signal": {},
        }

        image_list = [
            {
                "data": data,
                "axes": axes,
                "metadata": metadata,
                "original_metadata": header,
            }
        ]
        if read_em_image and has_em_image:
            image_list.append(
                {
                    "data": em_data,
                    "axes": axes_em,
                    "metadata": metadata_em,
                    "original_metadata": header,
                },
            )
        return image_list


def _parsejeol(fd):
    final_dict = {}
    tmp_list = []
    tmp_dict = final_dict

    mark = 1
    while abs(mark) == 1:
        mark = np.fromfile(fd, "b", 1)[0]
        if mark == 1:
            str_len = np.fromfile(fd, "<i", 1)[0]
            kwrd = fd.read(str_len).rstrip(b"\x00")

            if (
                kwrd == b"\xce\xdf\xb0\xc4"
            ):  # correct variable name which might be 'Port'
                kwrd = "Port"
            elif (
                kwrd[-1] == 222
            ):  # remove undecodable byte at the end of first ScanSize variable
                kwrd = _decode(kwrd[:-1])
            else:
                kwrd = _decode(kwrd)
            val_type, val_len = np.fromfile(fd, "<i", 2)
            tmp_list.append(kwrd)
            if val_type == 0:
                tmp_dict[kwrd] = {}
            else:
                c_type = jTYPE[val_type]
                arr_len = val_len // np.dtype(c_type).itemsize
                if c_type == "c":
                    value = fd.read(val_len).rstrip(b"\x00")
                    value = _decode(value).split("\x00")
                    # value = os.path.normpath(value.replace('\\','/')).split('\x00')
                else:
                    value = np.fromfile(fd, c_type, arr_len)
                if len(value) == 1:
                    value = value[0]
                if kwrd[-5:-1] == "PAGE":
                    kwrd = kwrd + "_" + value
                    tmp_dict[kwrd] = {}
                    tmp_list[-1] = kwrd
                elif kwrd == "CountRate" or kwrd == "DeadTime":
                    tmp_dict[kwrd] = {}
                    tmp_dict[kwrd]["value"] = value
                elif kwrd == "Limits":
                    pass
                    # see https://github.com/hyperspy/hyperspy/pull/2488
                    # first 16 bytes are encode in float32 and looks like limit values ([20. , 1., 2000, 1.] or [1., 0., 1000., 0.001])
                    # next 4 bytes are ascii character and looks like number format (%.0f or %.3f)
                    # next 12 bytes are unclear
                    # next 4 bytes are ascii character and are units (kV or nA)
                    # last 12 byes are unclear
                elif val_type == 14:
                    tmp_dict[kwrd] = {}
                    tmp_dict[kwrd]["index"] = value
                else:
                    tmp_dict[kwrd] = value
            if kwrd == "Limits":
                pass
                # see https://github.com/hyperspy/hyperspy/pull/2488
                # first 16 bytes are encode in int32 and looks like limit values (10, 1, 100000000, 1)
                # next 4 bytes are ascii character and looks like number format (%d)
                # next 12 bytes are unclear
                # next 4 bytes are ascii character and are units (mag)
                # last 12 byes are again unclear
            else:
                tmp_dict = tmp_dict[kwrd]
        else:
            if len(tmp_list) != 0:
                del tmp_list[-1]
                tmp_dict = final_dict
                for k in tmp_list:
                    tmp_dict = tmp_dict[k]
            else:
                mark = 0

    return final_dict


def _readcube(
    rawdata,
    frame_start_index,
    frame_list,
    width,
    height,
    channel_number,
    width_norm,
    height_norm,
    rebin_energy,
    SI_dtype,
    sweep,
    frame_shifts,
    sum_frames,
    read_em_image,
    only_valid_data,
    lazy,
):  # pragma: no cover
    """
        Read spectrum image (and SEM/STEM image) from pts file

        Parameters
        ----------
        rawdata : numpy.ndarray
            Spectrum image part of pts file.
        frame_start_index : np.ndarray of shape (sweep+1, ) or (0, )
            The indices of each frame start. If length is zero, the indices will be
            determined from rawdata.
        frame_list : list
            List of frames to be read.
        width, height : int
            The navigation dimension.
        channel_number : int
            The number of channels.
        width_norm, height_norm : int
            Rebin factor of the navigation dimension.
        rebin_energy : int
            Rebin factor of the energy dimension.
        sweep : int
            Number of sweep
        frame_shifts : list
            The list of image positions [[x0,y0,z0], ...]. The x, y, z values can
            be negative. The data points outside data cube are ignored.

        Returns
        -------
        data : numpy.ndarray or dask.array
            The spectrum image with shape (frame, x, y, energy) if sum_frames is
            False, otherwise (x, y, energy).
            If lazy is True, the dask array is a COO sparse array.
        em_data : numpy.ndarray or dask.array
            The SEM/STEM image with shape (frame, x, y) if sum_frames is False,
            otherwise (x, y).
        has_em_image : bool
            True if the stream contains SEM/STEM images.
        sweep : int
                The number of loaded frames.
        frame_start_index : list
            The indices of each frame start.
        max_shift : numpy.ndarray
            The maximum shifts of the origin in the navigation dimension.
    ```
        frame_shifts : numpy.ndarray
            The shifts of the origin in the navigation dimension for each frame.
    """

    # In case of sum_frames, spectrum image and SEM/STEM image are summing up to the same frame number.
    # To avoid overflow on integration of SEM/STEM image, data type of np.uint32 is selected
    # for 16 frames and over. (range of image intensity in each frame is 0-4095 (0-0xfff))
    EM_dtype = np.uint16
    frame_step = 1
    if sum_frames:
        frame_step = 0
        if sweep >= 16:
            EM_dtype = np.uint32
        n_frames = 1
    else:
        n_frames = sweep + 1

    if lazy:
        hypermap = np.zeros((n_frames), dtype=EM_dtype)  # dummy variable, not used
        data_list = []
    else:
        hypermap = np.zeros((n_frames, height, width, channel_number), dtype=SI_dtype)

    em_image = np.zeros((n_frames, width, height), dtype=EM_dtype)

    max_value = np.iinfo(SI_dtype).max

    frame_shifts = np.asarray(frame_shifts)
    frame_list = np.asarray(frame_list)
    max_shift = frame_shifts[frame_list].max(axis=0)
    min_shift = frame_shifts[frame_list].min(axis=0)
    #    sxyz = np.array([min_shift[0]-max_shift[0], min_shift[1]-max_shift[1],0])
    min_shift[2] = 0
    max_shift[2] = 0
    sxyz = min_shift - max_shift
    frame_shifts -= max_shift
    width += sxyz[1]
    height += sxyz[0]
    valid = None

    if lazy:
        readframe = _readframe_lazy
    else:
        readframe = _readframe_dense

    frame_num = 0
    p_start = 0
    target_frame_num = 0
    has_em_image = False
    for frame_idx in frame_list:
        if frame_idx < 0:  # skip invalid frame number
            continue

        # find the last indexed frame
        _i = frame_idx
        while _i > 0 and frame_start_index[_i] < 0:
            _i -= 1
        p_start = frame_start_index[_i]
        frame_num = _i

        while frame_num < frame_idx:
            # record start point of frame and skip frame
            p_start += _skip_frame(rawdata[p_start:])
            frame_num += 1
            frame_start_index[frame_num] = p_start

        if frame_idx < frame_shifts.size:
            fs = frame_shifts[frame_idx]
        else:
            fs = np.zeros(3, np.uint16)
            _logger.info(
                f"Size of frame_shift array is too small. The frame {frame_idx} is not moved."
            )
        length, frame_data, has_em, valid, max_valid = readframe(
            rawdata[p_start:],
            1,
            hypermap[target_frame_num],
            em_image[target_frame_num],
            width,
            height,
            channel_number,
            width_norm,
            height_norm,
            rebin_energy,
            fs[1],
            fs[0],
            fs[2],
            max_value,
        )
        has_em_image = has_em_image or has_em
        if length == 0:  # no data
            break
        if valid or not only_valid_data:
            # accept last frame
            if lazy:
                data_list.append(frame_data)
            frame_num += 1
            target_frame_num += frame_step
        else:
            # incomplete data, not accepted
            if sum_frames:
                # subtract signal counts of last frame
                _ = readframe(
                    rawdata[p_start:],
                    -1,
                    hypermap[target_frame_num],
                    em_image[target_frame_num],
                    width,
                    height,
                    channel_number,
                    width_norm,
                    height_norm,
                    rebin_energy,
                    fs[1],
                    fs[0],
                    fs[2],
                    max_value,
                )
                _logger.info(
                    "The last frame (sweep) is incomplete because the acquisition stopped during this frame. The partially acquired frame is ignored. Use 'sum_frames=False, only_valid_data=False' to read all frames individually, including the last partially completed frame."
                )
            break

        p_start += length
        if frame_num < frame_start_index.size:
            frame_start_index[frame_num] = p_start

    if not lazy:
        if sum_frames:
            # the first frame has integrated intensity
            return (
                hypermap[0, :height, :width],
                em_image[0, :height, :width],
                has_em_image,
                frame_num,
                frame_start_index,
                valid,
                max_shift,
                frame_shifts,
            )
        else:
            return (
                hypermap[:target_frame_num, :height, :width],
                em_image[:target_frame_num, :height, :width],
                has_em_image,
                frame_num,
                frame_start_index,
                valid,
                max_shift,
                frame_shifts,
            )

    # for lazy loading
    from rsciio.utils._fei_stream_readers import DenseSliceCOO

    length = np.sum([len(d) for d in data_list])
    # length = number of data points

    # v : [[frame_no, y, x, energy_channel, 1], ....]
    v = np.zeros(shape=(5, length), dtype=np.uint16)
    ptr = 0
    frame_count = 0
    for d in data_list:
        # d : data points in one frame
        d = np.asarray(d)
        # check if the pixels are in the valid data cube
        # (frame_shifts make partially integrated area at the rim)
        valid_cube = np.where(
            (0 <= d[:, 0])
            & (d[:, 0] < height)
            & (0 <= d[:, 1])
            & (d[:, 1] < width)
            & (0 <= d[:, 2])
            & (d[:, 2] < channel_number)
        )
        d = d[valid_cube]
        flen = len(d)
        pv = v[:, ptr : ptr + flen]
        pv[1:4, :] = np.array(d).transpose()
        pv[0, :] = frame_count
        pv[4, :] = 1
        ptr += flen
        frame_count += 1
    if sum_frames:
        data_shape = [height, width, channel_number]
        ar_s = DenseSliceCOO(v[1:4], v[4], shape=data_shape)
    else:
        data_shape = [frame_count, height, width, channel_number]
        ar_s = DenseSliceCOO(v[0:4], v[4], shape=data_shape)
    if sum_frames:
        em_image = em_image[0]

    import dask.array as da

    return (
        da.from_array(ar_s, asarray=False),
        em_image,
        has_em_image,
        sweep,
        frame_start_index,
        valid,
        max_shift,
        frame_shifts,
    )


@jit_ifnumba(cache=True)
def _readframe_dense(
    rawdata,
    countup,
    hypermap,
    em_image,
    width,
    height,
    channel_number,
    width_norm,
    height_norm,
    rebin_energy,
    dx,
    dy,
    dz,
    max_value,
):  # pragma: no cover
    """
    Read one frame from pts file. Used in a inner loop of _readcube function.
    This function always read SEM/STEM image even if read_em_image == False
    hypermap and em_image array will be modified

    Parameters
    ----------
    rawdata : numpy.ndarray of uint16
        slice of one frame raw data from whole raw data
    countup : 1 for summing up the X-ray events, -1 to cancel selected frame
    hypermap : numpy.ndarray(width, height, channel_number)
        numpy.ndarray to store decoded spectrum image.
    em_image : numpy.ndarray
        numpy.ndarray to store decoded SEM/TEM image with dimension (width, height)
        and dtype np.uint16 or np.uint32
    width : int
    height : int
    channel_number : int
        Limit of channel to reduce data size
    width_norm : int
        scanning step
    height_norm : int
        scanning step
    rebin_energy : int
        Binning parameter along energy axis. Must be 2^n.
    dx, dy, dz : int
        information of frame shift for drift correction.
    max_value : int
        limit of the data type used in hypermap array

    Returns
    -------
    raw_length : int
        frame length based on raw data array
    _ : int
        dummy value (used for lazy loading)
    has_em_image : bool
        True if pts file have SEM/STEM image
    valid : bool
        True if current frame is completely swept
        False for incomplete frame such as interrupt of measurement
    max_valid : int
        maximum number of scan lines to be accepted as valid
        in case of incomplete frame
    """

    count = 0
    has_em_image = False
    valid = False
    MAX_VAL = 4096
    previous_y = -1
    x = 0
    y = 0
    for value in rawdata:
        value_type = value & 0xF000
        value &= 0xFFF
        if value_type == 0x8000:
            x = value // width_norm + dx
            if x >= width:
                x = -1
            previous_x = value
        elif value_type == 0x9000:
            y = value // height_norm + dy
            if value < previous_y:
                break
            previous_y = value
            if y >= height:
                y = -1
        elif value_type == 0xA000 and x >= 0 and y >= 0:
            em_image[y, x] += value * countup
            has_em_image = True
        elif value_type == 0xB000:
            z = value // rebin_energy + dz
            if z < channel_number and x >= 0 and y >= 0 and z >= 0:
                hypermap[y, x, z] += countup
                if hypermap[y, x, z] == max_value:
                    raise ValueError(
                        "The range of the dtype is too small, "
                        "use `SI_dtype` to set a dtype with "
                        "higher range."
                    )
        count += 1

    if previous_y >= MAX_VAL - height_norm and (
        previous_x >= MAX_VAL - width_norm or not has_em_image
    ):
        # if the STEM-ADF/SEM-BF image is not included in pts file,
        # maximum value of x is usually smaller than max_x,
        # (sometimes no x-ray events occur in a last few pixels.)
        # So, only the y value is checked
        #
        # > is need when downsampling is specified
        valid = True
    return count, 0, has_em_image, valid, previous_y // height_norm


@jit_ifnumba(cache=True)
def _readframe_lazy(
    rawdata,
    _1,
    _2,
    em_image,
    width,
    height,
    channel_number,
    width_norm,
    height_norm,
    rebin_energy,
    dx,
    dy,
    dz,
    _3,
):  # pragma: no cover
    """
    Read one frame from pts file. Used in a inner loop of _readcube function.
    This function always read SEM/STEM image even if read_em_image == False
    hypermap and em_image array will be modified

    Parameters
    ----------
    rawdata : numpy.ndarray of uint16
        slice of one frame raw data from whole raw data
    _1 : dummy parameter, not used
    _2 : dummy parameter, not used
    em_image : numpy.ndarray(width, height), dtype = np.uint16 or np.uint32
        numpy.ndarray to store decoded SEM/TEM image.
    width : int
    height : int
    channel_number : int
        Limit of channel to reduce data size
    width_norm : int
        scanning step
    height_norm : int
        scanning step
    rebin_energy : int
        Binning parameter along energy axis. Must be 2^n.
    dx, dy, dz : int
        information of frame shift for drift correction.
    _3 : dummy parameter, not used


    Returns
    -------
    raw_length : int
        frame length based on raw data array
    data : list of [int, int, int]
        list of X-ray events as an array of [x, y, energy_ch]
    has_em_image : bool
        True if pts file have SEM/STEM image
    valid : bool
        True if current frame is completely swept
        False for incomplete frame such as interrupt of measurement
    max_valid : int
        maximum number of scan lines to be accepted as valid
        in case of incomplete frame
    """
    data = []
    previous_x = 0
    previous_y = 0
    MAX_VAL = 4096
    count = 0
    valid = False
    has_em_image = False
    for value in rawdata:
        count += 1
        value_type = value & 0xF000
        value &= 0xFFF
        if value_type == 0x8000:
            previous_x = value
            x = value // width_norm + dx
            if x >= width:
                x = -1
        elif value_type == 0x9000:
            if value < previous_y:
                break
            previous_y = value
            y = value // height_norm + dy
            if y >= width:
                y = -1
        elif value_type == 0xA000 and x >= 0 and y >= 0:
            em_image[y, x] += value
            has_em_image = True
        elif value_type == 0xB000:  # spectrum image
            z = value // rebin_energy + dz
            if 0 <= z and z < channel_number:
                data.append([y, x, z])
    if previous_y == MAX_VAL - height_norm and (
        previous_x == MAX_VAL - width_norm or not has_em_image
    ):
        # if the STEM-ADF/SEM-BF image is not included in pts file,
        # maximum value of x is usually smaller than max_x,
        # (sometimes no x-ray events occur in a last few pixels.)
        # So, only the y value is checked
        valid = True
    return count, data, has_em_image, valid, previous_y // height_norm


@jit_ifnumba(cache=True)
def _skip_frame(rawdata):  # pragma: no cover
    count = 0
    previous_y = 0
    for value in rawdata:
        value_type = value & 0xF000
        if value_type == 0x9000:
            y = value & 0xFFF
            if y < previous_y:
                break
            previous_y = y
        count += 1
    return count


def _read_eds(filename, **kwargs):
    """
    Parameters
    ----------
    filename : str
        path of .eds file

    Returns
    -------
    image_list : list of 1D signal (dict)
        list of dictionaries used to create the signals
        Always len(image_list) == 1
        None on read error
    """
    header = {}
    fd = open(filename, "br")
    # file_magic
    _ = np.fromfile(fd, "<I", 1)[0]
    np.fromfile(fd, "<b", 6)
    header["filedate"] = datetime(1899, 12, 30) + timedelta(
        days=np.fromfile(fd, "<d", 1)[0]
    )
    header["sp_name"] = _decode(fd.read(80).rstrip(b"\x00"))
    header["username"] = _decode(fd.read(32).rstrip(b"\x00"))

    np.fromfile(fd, "<i", 1)  # 1
    header["arr"] = np.fromfile(fd, "<d", 10)

    np.fromfile(fd, "<i", 1)  # 7
    np.fromfile(fd, "<d", 1)[0]
    header["Esc"] = np.fromfile(fd, "<d", 1)[0]
    header["Fnano F"] = np.fromfile(fd, "<d", 1)[0]
    header["E Noise"] = np.fromfile(fd, "<d", 1)[0]
    header["CH Res"] = np.fromfile(fd, "<d", 1)[0]
    header["live time"] = np.fromfile(fd, "<d", 1)[0]
    header["real time"] = np.fromfile(fd, "<d", 1)[0]
    header["DeadTime"] = np.fromfile(fd, "<d", 1)[0]
    header["CountRate"] = np.fromfile(fd, "<d", 1)[0]
    header["CountRate n"] = np.fromfile(fd, "<i", 1)[0]
    header["CountRate sum"] = np.fromfile(fd, "<d", 2)
    header["CountRate value"] = np.fromfile(fd, "<d", 1)[0]
    np.fromfile(fd, "<d", 1)[0]
    header["DeadTime n"] = np.fromfile(fd, "<i", 1)[0]
    header["DeadTime sum"] = np.fromfile(fd, "<d", 2)
    header["DeadTime value"] = np.fromfile(fd, "<d", 1)[0]
    np.fromfile(fd, "<d", 1)[0]
    header["CoefA"] = np.fromfile(fd, "<d", 1)[0]
    header["CoefB"] = np.fromfile(fd, "<d", 1)[0]
    header["State"] = _decode(fd.read(32).rstrip(b"\x00"))
    np.fromfile(fd, "<i", 1)[0]
    np.fromfile(fd, "<d", 1)[0]
    header["Tpl"] = _decode(fd.read(32).rstrip(b"\x00"))
    header["NumCH"] = np.fromfile(fd, "<i", 1)[0]
    data = np.fromfile(fd, "<i", header["NumCH"])

    footer = {}
    _ = np.fromfile(fd, "<i", 1)

    n_fbd_elem = np.fromfile(fd, "<i", 1)[0]
    if n_fbd_elem != 0:
        list_fbd_elem = np.fromfile(fd, "<H", n_fbd_elem)
        footer["Excluded elements"] = list_fbd_elem

    n_elem = np.fromfile(fd, "<i", 1)[0]
    if n_elem != 0:
        elems = {}
        for i in range(n_elem):
            # mark elem
            _ = np.fromfile(fd, "<i", 1)[0]  # = 2
            # Z
            _ = np.fromfile(fd, "<H", 1)[0]
            # mark1, mark2
            _, _ = np.fromfile(fd, "<i", 2)  # = 1, 0
            roi_min, roi_max = np.fromfile(fd, "<H", 2)
            # unknown
            _ = np.fromfile(fd, "<b", 14)
            energy, unknow1, unknow2, unknow3 = np.fromfile(fd, "<d", 4)
            elem_name = _decode(fd.read(32).rstrip(b"\x00"))
            # mark3?
            _ = np.fromfile(fd, "<i", 1)[0]
            n_line = np.fromfile(fd, "<i", 1)[0]
            lines = {}
            for j in range(n_line):
                # mark_line?
                _ = np.fromfile(fd, "<i", 1)[0]  # = 1
                e_line = np.fromfile(fd, "<d", 1)[0]
                z = np.fromfile(fd, "<H", 1)[0]
                e_length = np.fromfile(fd, "<b", 1)[0]
                e_name = _decode(fd.read(e_length).rstrip(b"\x00"))
                l_length = np.fromfile(fd, "<b", 1)[0]
                l_name = _decode(fd.read(l_length).rstrip(b"\x00"))
                detect = np.fromfile(fd, "<i", 1)[0]
                lines[e_name + "_" + l_name] = {
                    "energy": e_line,
                    "Z": z,
                    "detection": detect,
                }
            elems[elem_name] = {
                "Z": z,
                "Roi_min": roi_min,
                "Roi_max": roi_max,
                "Energy": energy,
                "Lines": lines,
            }
        footer["Selected elements"] = elems

    n_quanti = np.fromfile(fd, "<i", 1)[0]
    if n_quanti != 0:
        # all unknown
        _ = np.fromfile(fd, "<i", 1)[0]
        _ = np.fromfile(fd, "<i", 1)[0]
        _ = np.fromfile(fd, "<i", 1)[0]
        _ = np.fromfile(fd, "<d", 1)[0]
        _ = np.fromfile(fd, "<i", 1)[0]
        _ = np.fromfile(fd, "<i", 1)[0]
        quanti = {}
        for i in range(n_quanti):
            # mark elem
            _ = np.fromfile(fd, "<i", 1)[0]  # = 2
            z = np.fromfile(fd, "<H", 1)[0]
            # mark1, mark2
            _, _ = np.fromfile(fd, "<i", 2)  # = 1, 0
            energy, unkn6 = np.fromfile(fd, "<d", 2)
            mass1 = np.fromfile(fd, "<d", 1)[0]
            error = np.fromfile(fd, "<d", 1)[0]
            atom = np.fromfile(fd, "<d", 1)[0]
            ox_name = _decode(fd.read(16).rstrip(b"\x00"))
            mass2 = np.fromfile(fd, "<d", 1)[0]
            # K
            _ = np.fromfile(fd, "<d", 1)[0]
            counts = np.fromfile(fd, "<d", 1)[0]
            # all unknown
            _ = np.fromfile(fd, "<d", 1)[0]
            _ = np.fromfile(fd, "<d", 1)[0]
            _ = np.fromfile(fd, "<i", 1)[0]
            _ = np.fromfile(fd, "<i", 1)[0]
            _ = np.fromfile(fd, "<d", 1)[0]
            quanti[ox_name] = {
                "Z": z,
                "Mass1 (%)": mass1,
                "Error": error,
                "Atom (%)": atom,
                "Mass2 (%)": mass2,
                "Counts": counts,
                "Energy": energy,
            }
        footer["Quanti"] = quanti

    e = np.fromfile(fd, "<i", 1)
    if e == 5:
        footer["Parameters"] = {
            "DetT": _decode(fd.read(16).rstrip(b"\x00")),
            "SEM": _decode(fd.read(16).rstrip(b"\x00")),
            "Port": _decode(fd.read(16).rstrip(b"\x00")),
            "AccKV": np.fromfile(fd, "<d", 1)[0],
            "AccNA": np.fromfile(fd, "<d", 1)[0],
            "skip": np.fromfile(fd, "<b", 38),
            "MnKaRES": np.fromfile(fd, "d", 1)[0],
            "WorkD": np.fromfile(fd, "d", 1)[0],  # codespell:ignore
            "InsD": np.fromfile(fd, "d", 1)[0],
            "XtiltAng": np.fromfile(fd, "d", 1)[0],
            "TakeAng": np.fromfile(fd, "d", 1)[0],
            "IncAng": np.fromfile(fd, "d", 1)[0],
            "skip2": np.fromfile(fd, "<i", 1)[0],
            "ScanSize": np.fromfile(fd, "d", 1)[0],
            "DT_64": np.fromfile(fd, "<H", 1)[0],
            "DT_128": np.fromfile(fd, "<H", 1)[0],
            "DT_256": np.fromfile(fd, "<H", 1)[0],
            "DT_512": np.fromfile(fd, "<H", 1)[0],
            "DT_1K": np.fromfile(fd, "<H", 1)[0],
            "DetH": np.fromfile(fd, "d", 1)[0],
            "DirAng": np.fromfile(fd, "d", 1)[0],
            "XtalAng": np.fromfile(fd, "d", 1)[0],
            "ElevAng": np.fromfile(fd, "d", 1)[0],
            "ValidSize": np.fromfile(fd, "d", 1)[0],
            "WinCMat": _decode(fd.read(4).rstrip(b"\x00")),
            "WinCZ": np.fromfile(fd, "<H", 1)[0],
            "WinCThic": np.fromfile(fd, "d", 1)[0],
            "WinChem": _decode(fd.read(16).rstrip(b"\x00")),
            "WinChem_nelem": np.fromfile(fd, "<H", 1)[0],
            "WinChem_Z1": np.fromfile(fd, "<H", 1)[0],
            "WinChem_Z2": np.fromfile(fd, "<H", 1)[0],
            "WinChem_Z3": np.fromfile(fd, "<H", 1)[0],
            "WinChem_Z4": np.fromfile(fd, "<H", 1)[0],
            "WinChem_Z5": np.fromfile(fd, "<H", 1)[0],
            "WinChem_m1": np.fromfile(fd, "d", 1)[0],
            "WinChem_m2": np.fromfile(fd, "d", 1)[0],
            "WinChem_m3": np.fromfile(fd, "d", 1)[0],
            "WinChem_m4": np.fromfile(fd, "d", 1)[0],
            "WinChem_m5": np.fromfile(fd, "d", 1)[0],
            "WinThic": np.fromfile(fd, "d", 1)[0],
            "WinDens": np.fromfile(fd, "d", 1)[0],
            "SpatMat": _decode(fd.read(4).rstrip(b"\x00")),
            "SpatZ": np.fromfile(fd, "<H", 1)[0],
            "SpatThic": np.fromfile(fd, "d", 1)[0],
            "SiDead": np.fromfile(fd, "d", 1)[0],
            "SiThic": np.fromfile(fd, "d", 1)[0],
        }

    hv = footer["Parameters"]["AccKV"]
    if hv <= 30.0:
        mode = "SEM"
    else:
        mode = "TEM"

    axes = [
        {
            "name": "Energy",
            "size": header["NumCH"],
            "offset": header["CoefB"],
            "scale": header["CoefA"],
            "units": "keV",
            "navigate": False,
        }
    ]

    metadata = {
        "Acquisition_instrument": {
            mode: {
                "beam_energy": hv,
                "Detector": {
                    "EDS": {
                        "azimuth_angle": footer["Parameters"]["DirAng"],
                        "detector_type": footer["Parameters"]["DetT"],
                        "elevation_angle": footer["Parameters"]["ElevAng"],
                        "energy_resolution_MnKa": footer["Parameters"]["MnKaRES"],
                        "live_time": header["live time"],
                    },
                },
            },
        },
        "General": {
            "original_filename": os.path.basename(filename),
            "date": header["filedate"].date().isoformat(),
            "time": header["filedate"].time().isoformat(),
            "title": "EDX",
        },
        "Signal": {
            "quantity": "X-rays (Counts)",
            "signal_type": "EDS_" + mode,
        },
    }

    dictionary = {
        "data": data,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": {"Header": header, "Footer": footer},
    }

    return [dictionary]


extension_to_reader_mapping = {
    "img": _read_img,
    "map": _read_img,
    "pts": _read_pts,
    "eds": _read_eds,
    "asw": _read_asw,
}


def _decode(bytes_string):
    try:
        string = bytes_string.decode("utf-8")
    except Exception:
        # See https://github.com/hyperspy/hyperspy/issues/2812
        string = bytes_string.decode("shift_jis")

    return string



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# The EMD format is a hdf5 standard proposed at Lawrence Berkeley
# National Lab (see https://emdatasets.com/ for more information).
# FEI later developed another EMD format, also based on the hdf5 standard. This
# reader first checked if the file have been saved by Velox (FEI EMD format)
# and use either the EMD class or the FEIEMDReader class to read the file.
# Writing file is only supported for EMD Berkeley file.

import json
import logging
from pathlib import Path

import h5py

from rsciio._docstrings import (
    CHUNKS_DOC,
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SIGNAL_DOC,
)

from ._emd_ncem import read_emd_version

_logger = logging.getLogger(__name__)


def is_EMD_NCEM(file):
    """
    Parameters
    ----------
    file : h5py file handle
        DESCRIPTION.

    Returns
    -------
    bool
        DESCRIPTION.

    """

    def _is_EMD_NCEM(file):
        # the version can be defined in the root or the group
        if read_emd_version(file) != "":
            return True
        for key in file.keys():
            item = file[key]
            if isinstance(item, h5py.Group) and read_emd_version(item) != "":
                return True
        return False

    if isinstance(file, str):
        with h5py.File(file, "r") as f:
            return _is_EMD_NCEM(f)
    else:
        return _is_EMD_NCEM(file)


def is_EMD_Velox(file):
    """Function to check if the EMD file is an Velox file.

    Parameters
    ----------
    file : string or HDF5 file handle
        The name of the emd-file from which to load the signals. Standard
        file extension is 'emd'.

    Returns
    -------
    True if the file is a Velox file, otherwise False

    """

    def _is_EMD_velox(file):
        if "Version" in list(file.keys()):
            version = file.get("Version")
            v_dict = json.loads(version[0].decode("utf-8"))
            if v_dict["format"] in ["Velox", "DevelopersKit"]:
                return True
        return False

    if isinstance(file, Path):
        file = str(file)

    if isinstance(file, str):
        with h5py.File(file, "r") as f:
            return _is_EMD_velox(f)
    else:
        return _is_EMD_velox(file)


def file_reader(
    filename,
    lazy=False,
    dataset_path=None,
    stack_group=None,
    select_type=None,
    first_frame=0,
    last_frame=None,
    sum_frames=True,
    sum_EDS_detectors=True,
    rebin_energy=1,
    SI_dtype=None,
    load_SI_image_stack=False,
):
    """
    Read EMD file, which can be an NCEM or a Velox variant of the EMD format.
    Also reads Direct Electron's DE5 format, which is read according to the
    NCEM specifications.

    Parameters
    ----------
    %s
    %s
    dataset_path : None, str or list of str, default=None
        NCEM only: Path of the dataset. If None, load all supported datasets,
        otherwise the specified dataset(s).
    stack_group : None, bool, default=None
        NCEM only: Stack datasets of groups with common path. Relevant for emd file
        version >= 0.5, where groups can be named ``group0000``, ``group0001``, etc.
        For py4DSTEM files saved by EMD version 1.0, set ``stack_group=False`` to load
        the 4D datacube as well as virtual images, dp_mean, etc.
    select_type : {None, 'image', 'single_spectrum', 'spectrum_image'}
        Velox only: specifies the type of data to load: if ``'image'`` is selected,
        only images (including EDS maps) are loaded, if ``'single_spectrum'`` is
        selected, only single spectra are loaded and if ``'spectrum_image'`` is
        selected, only the spectrum image will be loaded.
    first_frame : int, default=0
        Velox only: Select the start for the frame range of the EDS spectrum image
        to load.
    last_frame : int or None, default=None
        Velox only: Select the end for the frame range of the EDS spectrum image
        to load.
    sum_frames : bool, default=True
        Velox only: Load each individual EDS frame. The EDS spectrum image will
        be loaded with an extra navigation dimension corresponding to the frame
        index (time axis).
    sum_EDS_detectors : bool, default=True
        Velox only: Load the EDS signal as a sum over the signals from all EDS
        detectors (default) or, alternatively, load the signal of each individual
        EDS detector. In the latter case, a corresponding number of distinct
        EDS signals is returned.
    rebin_energy : int, default=1
        Velox only: Rebin the energy axis by given factor. Useful in combination
        with ``sum_frames=False`` to reduce the data size when reading the
        individual frames of the spectrum image.
    SI_dtype : numpy.dtype or None, default=None
        Velox only: Change the datatype of a spectrum image. Useful in combination
        with ``sum_frames=False`` to reduce the data size when reading the individual
        frames of the spectrum image. If ``None``, the dtype of the data in
        the emd file is used.
    load_SI_image_stack : bool, default=False
        Velox only: Allows loading the stack of STEM images acquired
        simultaneously with the EDS spectrum image. This option can be useful to
        monitor any specimen changes during the acquisition or to correct the
        spatial drift in the spectrum image by using the STEM images.

    %s

    Raises
    ------
    ModuleNotFoundError
        When reading spectrum image from Velox EMD file and the ``sparse`` library is missing.
    """
    if isinstance(filename, Path):
        filename = str(filename)

    file = h5py.File(filename, "r")
    dictionaries = []
    try:
        if is_EMD_Velox(file):
            from ._emd_velox import FeiEMDReader

            _logger.debug("EMD file is a Velox variant.")
            emd_reader = FeiEMDReader(
                lazy=lazy,
                select_type=select_type,
                first_frame=first_frame,
                last_frame=last_frame,
                sum_frames=sum_frames,
                sum_EDS_detectors=sum_EDS_detectors,
                rebin_energy=rebin_energy,
                SI_dtype=SI_dtype,
                load_SI_image_stack=load_SI_image_stack,
            )
            emd_reader.read_file(file)
        elif is_EMD_NCEM(file):
            from ._emd_ncem import EMD_NCEM

            _logger.debug("EMD file is a Berkeley variant.")
            emd_reader = EMD_NCEM()
            emd_reader.read_file(
                file, lazy=lazy, dataset_path=dataset_path, stack_group=stack_group
            )
        else:
            raise IOError("The file is not a supported EMD file.")
    except Exception as e:
        raise e
    finally:
        if not lazy:
            file.close()

    dictionaries = emd_reader.dictionaries

    return dictionaries


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def file_writer(filename, signal, chunks=None, **kwds):
    """
    Write signal to EMD file. Only the specifications by the National Center
    for Electron Microscopy (NCEM) are supported.

    Parameters
    ----------
    %s
    %s
    %s
    **kwds : dict, optional
        Dictionary containing metadata, which will be written as attribute
        of the root group.
    """
    if isinstance(filename, Path):
        filename = str(filename)
    from ._emd_ncem import EMD_NCEM

    EMD_NCEM().write_file(filename, signal, chunks=chunks, **kwds)


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    CHUNKS_DOC,
)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
import os

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC

_logger = logging.getLogger(__name__)


try:
    from netCDF4 import Dataset as netcdf_file_reader

    netcdf_reader = "netCDF4"
except Exception:
    try:
        from scipy.io import netcdf_file as netcdf_file_reader

        netcdf_reader = "scipy"
    except Exception:
        netcdf_reader = None


attrib2netcdf = {
    "energyorigin": "energy_origin",
    "energyscale": "energy_scale",
    "energyunits": "energy_units",
    "xorigin": "x_origin",
    "xscale": "x_scale",
    "xunits": "x_units",
    "yorigin": "y_origin",
    "yscale": "y_scale",
    "yunits": "y_units",
    "zorigin": "z_origin",
    "zscale": "z_scale",
    "zunits": "z_units",
    "exposure": "exposure",
    "title": "title",
    "binning": "binning",
    "readout_frequency": "readout_frequency",
    "ccd_height": "ccd_height",
    "blanking": "blanking",
}

acquisition2netcdf = {
    "exposure": "exposure",
    "binning": "binning",
    "readout_frequency": "readout_frequency",
    "ccd_height": "ccd_height",
    "blanking": "blanking",
    "gain": "gain",
    "pppc": "pppc",
}

treatments2netcdf = {
    "dark_current": "dark_current",
    "readout": "readout",
}


def file_reader(filename, lazy=False):
    """
    Read netCDF ``.nc`` files saved using the HyperSpy predecessor EELSlab.

    Parameters
    ----------
    %s
    %s

    %s
    """
    if netcdf_reader is None:
        raise ImportError(
            "No netCDF library installed. "
            "To read EELSLab netcdf files install "
            "one of the following packages:"
            "netCDF4 or scipy."
        )

    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    ncfile = netcdf_file_reader(filename, "r")

    if (
        hasattr(ncfile, "file_format_version")
        and ncfile.file_format_version == "EELSLab 0.1"
    ):
        dictionary = nc_hyperspy_reader_0dot1(ncfile, filename)
    else:
        ncfile.close()
        raise IOError("Unsupported netCDF file")

    return (dictionary,)


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC)


def nc_hyperspy_reader_0dot1(ncfile, filename):
    calibration_dict, acquisition_dict, treatments_dict = {}, {}, {}
    dc = ncfile.variables["data_cube"]
    data = dc[:]
    if "history" in calibration_dict:
        calibration_dict["history"] = eval(ncfile.history)
    for attrib in attrib2netcdf.items():
        if hasattr(dc, attrib[1]):
            value = eval("dc." + attrib[1])
            if isinstance(value, np.ndarray):
                calibration_dict[attrib[0]] = value[0]
            else:
                calibration_dict[attrib[0]] = value
        else:
            _logger.warning(
                "Warning: the attribute '%s' is not defined in the file '%s'",
                attrib[0],
                filename,
            )
    for attrib in acquisition2netcdf.items():
        if hasattr(dc, attrib[1]):
            value = eval("dc." + attrib[1])
            if isinstance(value, np.ndarray):
                acquisition_dict[attrib[0]] = value[0]
            else:
                acquisition_dict[attrib[0]] = value
        else:
            _logger.warning(
                "Warning: the attribute '%s' is not defined in the file '%s'",
                attrib[0],
                filename,
            )
    for attrib in treatments2netcdf.items():
        if hasattr(dc, attrib[1]):
            treatments_dict[attrib[0]] = eval("dc." + attrib[1])
        else:
            _logger.warning(
                "Warning: the attribute '%s' is not defined in the file '%s'",
                attrib[0],
                filename,
            )
    original_metadata = {
        "record_by": ncfile.type,
        "calibration": calibration_dict,
        "acquisition": acquisition_dict,
        "treatments": treatments_dict,
    }
    ncfile.close()
    # Now we'll map some parameters
    record_by = "image" if original_metadata["record_by"] == "image" else "spectrum"
    if record_by == "image":
        dim = len(data.shape)
        names = ["Z", "Y", "X"][3 - dim :]
        scaleskeys = ["zscale", "yscale", "xscale"]
        originskeys = ["zorigin", "yorigin", "xorigin"]
        unitskeys = ["zunits", "yunits", "xunits"]
        navigate = [True, False, False]

    elif record_by == "spectrum":
        dim = len(data.shape)
        names = ["Y", "X", "Energy"][3 - dim :]
        scaleskeys = ["yscale", "xscale", "energyscale"]
        originskeys = ["yorigin", "xorigin", "energyorigin"]
        unitskeys = ["yunits", "xunits", "energyunits"]
        navigate = [True, True, False]

    # The images are recorded in the Fortran order
    data = data.T.copy()
    try:
        scales = [calibration_dict[key] for key in scaleskeys[3 - dim :]]
    except KeyError:
        scales = [1, 1, 1][3 - dim :]
    try:
        origins = [calibration_dict[key] for key in originskeys[3 - dim :]]
    except KeyError:
        origins = [0, 0, 0][3 - dim :]
    try:
        units = [calibration_dict[key] for key in unitskeys[3 - dim :]]
    except KeyError:
        units = ["", "", ""]
    axes = [
        {
            "size": int(data.shape[i]),
            "index_in_array": i,
            "name": names[i],
            "scale": scales[i],
            "offset": origins[i],
            "units": units[i],
            "navigate": navigate[i],
        }
        for i in range(dim)
    ]
    metadata = {"General": {}, "Signal": {}}
    metadata["General"]["original_filename"] = os.path.split(filename)[1]
    metadata["General"]["signal_type"] = ""
    dictionary = {
        "data": data,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": original_metadata,
    }

    return dictionary



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# INFORMATION ABOUT THE SEMPER FORMAT LABEL:
# Picture labels consist of a sequence of bytes
# ---------------------------------------------
# v61:256B | v7:at least 256B, rounded up to multiple of block size 64,
#               with max set by LNLAB in params.f
# The versions have the same contents for the first 256B, as set out below,
# and referred to as the label 'base'; beyond this, in the label 'extension',
# the structure is deliberately undefined, allowing you to make your own use
# of the additional storage.
#
# From Aug14:
#  B1-6    S e m p e r (ic chars)
#   7,8    ncol msb,lsb (BIG-ended)
#   9,10   nrow msb,lsb
#  11,12   nlay msb,lsb
#  13,14   ccol msb,lsb
#  15,16   crow msb,lsb
#  17,18   clay msb,lsb
#    19    class: 1-20
#            for image,macro,fou,spec,correln,undef,walsh,histog,plist,lut
#    20    form: 0,1,2,3,4 = byte,i*2,fp,com,i*4 from Aug08
#    21    wp: non-zero if prot
#  22-27   creation year(-1900?),month,day,hour,min,sec
# 28    v61|v7 # chars in range record | 255
#  29-55   v61: min,max values present (ic chars for decimal repn)
#           v7: min,max vals as two Fp values in B29-36 (LE ordered)
#               followed by 19 unused bytes B37-55
#    56    plist type: 1,2,3 = list,opencurve,closedcurve
#            acc to EXAMPD - code appears to use explicit numbers
# 57,58,59 ncol,nrow,nlay hsb
# 60,61,62 ccol,crow,clay hsb
#    63    RealCoords flag (non-zero -> DX,DY,DZ,X0,Y0,Z0,units held as below)
# 64    v61:0 | v7: # blocks in (this) picture label (incl extn)
#  65-67   0 (free at present)
#  68-71   DATA cmd V7   4-byte Fp values, order LITTLE-ended
#  72-75   DATA cmd V6
#  76-79   RealCoord DZ / V5  RealCoord pars as 4-byte Fp values, LITTLE-ended
#  80-83   ...       Z0 / V4
#  84-87   ...       DY / V3
#  88-91   ...       Y0 / V2
#  92-95   ...       DX / V1
#  96-99   ...       X0 / V0
# 100    # chars in title
# 101-244  title (ic chars)
# 245-248  RealCoord X unit (4 ic chars)
# 249-252  RealCoord Y unit (4 ic chars)
# 253-256  RealCoord Z unit (4 ic chars)
#
# Aug08-Aug14 labels held no RealCoord information, so:
#   B63    'free' and zero - flagging absence of RealCoord information
# 101-256  title (12 chars longer than now)
#
# Before Aug08 labels held no hsb for pic sizes, so:
#  57-99   all free/zero except for use by DATA cmd
# 101-256  title (ic chars)

import logging
import struct
import warnings
from collections import OrderedDict
from datetime import datetime
from functools import partial

import numpy as np

from rsciio._docstrings import (
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SIGNAL_DOC,
)
from rsciio.utils._array import sarray2dict
from rsciio.utils._dictionary import DTBox

_logger = logging.getLogger(__name__)


class SemperFormat(object):
    """Class for importing and exporting SEMPER `.unf`-files.

    The :class:`~.SemperFormat` class represents a SEMPER binary file format
    with a header, which holds additional information. `.unf`-files can be
    saved and read from files.

    Attributes
    ----------
    data : :class:`~numpy.ndarray` (N=3)
        The phase map or magnetization information in a 3D array (with one
        slice).
    title : string
        Title of the file (not to be confused with the filename).
    offsets : tuple (N=3) of floats
        Offset shifts (in nm) of the grid origin (does not have to start at 0)
        in x, y, z.
    scales : tuple (N=3) of floats
        Grid spacing (nm per pixel) in x, y, z.
    units : tuple (N=3) of strings
        Units of the grid in x, y, z.
    metadata : dictionary
        A dictionary of all flags and metadata present in the `.unf`-file.

    """

    ICLASS_DICT = {
        1: "image",
        2: "macro",
        3: "fourier",
        4: "spectrum",
        5: "correlation",
        6: None,
        7: "walsh",
        8: "position list",
        9: "histogram",
        10: "display look-up table",
    }

    ICLASS_DICT_INV = {v: k for k, v in ICLASS_DICT.items()}

    IFORM_DICT = {0: np.byte, 1: np.int16, 2: np.float32, 3: np.complex64, 4: np.int32}

    IFORM_DICT_INV = {v: k for k, v in IFORM_DICT.items()}

    HEADER_DTYPES = [
        ("NCOL", "<i2"),
        ("NROW", "<i2"),
        ("NLAY", "<i2"),
        ("ICLASS", "<i2"),
        ("IFORM", "<i2"),
        ("IFLAG", "<i2"),
        ("IFORM", "<i2"),
    ]

    LABEL_DTYPES = [
        ("SEMPER", ("<i2", 6)),  # Bytes 1-6
        ("NCOL", ("<i2", 2)),  # Bytes 7,8
        ("NROW", ("<i2", 2)),  # Bytes 9,10
        ("NLAY", ("<i2", 2)),  # Bytes 11,12
        ("ICCOLN", ("<i2", 2)),  # Bytes 13,14
        ("ICROWN", ("<i2", 2)),  # Bytes 15,16
        ("ICLAYN", ("<i2", 2)),  # Bytes 17,18
        ("ICLASS", "<i2"),  # Bytes 19
        ("IFORM", "<i2"),  # Bytes 20
        ("IWP", "<i2"),  # Bytes 21
        ("DATE", ("<i2", 6)),  # Bytes 22-27
        ("NCRANG", "<i2"),  # Bytes 28
        ("RANGE", ("<i2", 27)),  # Bytes 29-55
        ("IPLTYP", "<i2"),  # Bytes 56
        ("NCOLH", "<i2"),  # Bytes 57
        ("NROWH", "<i2"),  # Bytes 58
        ("NLAYH", "<i2"),  # Bytes 59
        ("ICCOLNH", "<i2"),  # Bytes 60
        ("ICROWNH", "<i2"),  # Bytes 61
        ("ICLAYNH", "<i2"),  # Bytes 62
        ("REALCO", "<i2"),  # Bytes 63
        ("NBLOCK", "<i2"),  # Bytes 64
        ("FREE", ("<i2", 3)),  # Bytes 65-67
        ("DATAV6", ("<i2", 4)),  # Bytes 68-71
        ("DATAV7", ("<i2", 4)),  # Bytes 72-75
        ("DZV5", ("<i2", 4)),  # Bytes 76-79
        ("Z0V4", ("<i2", 4)),  # Bytes 80-83
        ("DYV3", ("<i2", 4)),  # Bytes 84-87
        ("Y0V2", ("<i2", 4)),  # Bytes 88-91
        ("DXV1", ("<i2", 4)),  # Bytes 92-95
        ("X0V0", ("<i2", 4)),  # Bytes 96-99
        ("NTITLE", "<i2"),  # Bytes 100
        ("TITLE", ("<i2", 144)),  # Bytes 101-244
        ("XUNIT", ("<i2", 4)),  # Bytes 245-248
        ("YUNIT", ("<i2", 4)),  # Bytes 249-252
        ("ZUNIT", ("<i2", 4)),
    ]  # Bytes 253-256

    def __init__(
        self,
        data,
        title="",
        offsets=(0.0,) * 3,
        scales=(1.0,) * 3,
        units=(None,) * 3,
        metadata=None,
    ):
        if metadata is None:
            metadata = {}
        # Make sure data is 3D!
        data = data[tuple(None for _ in range(3 - len(data.shape)))]
        self.data = data
        self.title = title
        self.offsets = offsets
        self.scales = scales
        self.units = units
        self.metadata = metadata
        _logger.debug("Created " + str(self))

    @classmethod
    def _read_label(cls, unf_file):
        unpack = partial(
            unpack_from_intbytes, "<f"
        )  # Unpacking function for 4 byte floats!
        rec_length = np.fromfile(unf_file, dtype="<i", count=1)[0]  # length of label
        label = sarray2dict(np.fromfile(unf_file, dtype=cls.LABEL_DTYPES, count=1))
        label["SEMPER"] = "".join([str(chr(label_)) for label_ in label["SEMPER"]])
        assert label["SEMPER"] == "Semper"
        # Process dimensions:
        for key in ["NCOL", "NROW", "NLAY", "ICCOLN", "ICROWN", "ICLAYN"]:
            value = (
                256**2 * np.int32(label.pop(key + "H"))
                + 256 * label[key][0]
                + label[key][1]
            )
            label[key] = value
        # Process date:
        date = "{}-{}-{} {}:{}:{}".format(label["DATE"][0] + 1900, *label["DATE"][1:])
        label["DATE"] = date
        # Process range:
        if label["NCRANG"] == 255:
            range_min = unpack(label["RANGE"][:4])
            range_max = unpack(label["RANGE"][4:8])
            range_string = "{:.6g},{:.6g}".format(range_min, range_max)
        else:
            range_string = "".join(
                [str(chr(label_)) for label_ in label["RANGE"][: label["NCRANG"]]]
            )
        label["RANGE"] = range_string
        # Process real coords:
        x0 = unpack(label.pop("X0V0"))
        dx = unpack(label.pop("DXV1"))
        y0 = unpack(label.pop("Y0V2"))
        dy = unpack(label.pop("DYV3"))
        z0 = unpack(label.pop("Z0V4"))
        dz = unpack(label.pop("DZV5"))
        if label["REALCO"] == 1:
            label.update(
                {"X0V0": x0, "Y0V2": y0, "Z0V4": z0, "DXV1": dx, "DYV3": dy, "DZV5": dz}
            )
        # Process additional commands (unused, not sure about the purpose):
        data_v6 = unpack(label["DATAV6"])
        data_v7 = unpack(label["DATAV7"])
        label["DATAV6"] = data_v6
        label["DATAV7"] = data_v7
        # Process title:
        title = "".join(
            [str(chr(label_)) for label_ in label["TITLE"][: label["NTITLE"]]]
        )
        label["TITLE"] = title
        # Process units:
        label["XUNIT"] = "".join([chr(label_) for label_ in label["XUNIT"]]).replace(
            "\x00", ""
        )
        label["YUNIT"] = "".join([chr(label_) for label_ in label["YUNIT"]]).replace(
            "\x00", ""
        )
        label["ZUNIT"] = "".join([chr(label_) for label_ in label["ZUNIT"]]).replace(
            "\x00", ""
        )
        # Sanity check:
        assert np.fromfile(unf_file, dtype="<i4", count=1)[0] == rec_length
        # Return label:
        return label

    def _get_label(self):
        pack = partial(pack_to_intbytes, "<f")  # Packing function for 4 byte floats!
        nlay, nrow, ncol = self.data.shape
        data, iform = self._check_format(self.data)
        title = self.title
        # Create label:
        label = np.zeros((1,), dtype=self.LABEL_DTYPES)
        # Fill label:
        label["SEMPER"] = [ord(c) for c in "Semper"]
        label["NCOLH"], remain = divmod(ncol, 256**2)
        label["NCOL"] = divmod(remain, 256)
        label["NROWH"], remain = divmod(nrow, 256**2)
        label["NROW"] = divmod(remain, 256)
        label["NLAYH"], remain = divmod(nlay, 256**2)
        label["NLAY"] = divmod(remain, 256)
        iccoln = self.metadata.get("ICCOLN", self.data.shape[2] // 2 + 1)
        label["ICCOLNH"], remain = divmod(iccoln, 256**2)
        label["ICCOLN"] = divmod(remain, 256)
        icrown = self.metadata.get("ICROWN", self.data.shape[1] // 2 + 1)
        label["ICROWNH"], remain = divmod(icrown, 256**2)
        label["ICROWN"] = divmod(remain, 256)
        iclayn = self.metadata.get("ICLAYN", self.data.shape[0] // 2 + 1)
        label["ICLAYNH"], remain = divmod(iclayn, 256**2)
        label["ICLAYN"] = divmod(remain, 256)
        label["ICLASS"] = self.metadata.get("ICLASS", 6)  # 6: Undefined!
        label["IFORM"] = iform
        label["IWP"] = self.metadata.get("IWP", 0)  # seems standard
        date = self.metadata.get("DATE", "%s" % datetime.now())
        year, time = date.split(" ")
        date_ints = list(map(int, year.split("-"))) + list(map(int, time.split(":")))
        date_ints[0] -= 1900  # Modify year integer!
        label["DATE"] = date_ints
        range_string = "{:.4g},{:.4g}".format(self.data.min(), self.data.max())
        label["NCRANG"] = len(range_string)
        label["RANGE"][:, : len(range_string)] = [ord(c) for c in range_string]
        label["IPLTYP"] = self.metadata.get("IPLTYP", 248)  # seems standard
        label["REALCO"] = 1  # Real coordinates are used!
        label["NBLOCK"] = 4  # Always 4 64b blocks!
        label["FREE"] = [0, 0, 0]
        label["DATAV6"] = pack(0.0)  # Not used!
        label["DATAV7"] = pack(0.0)  # Not used!
        label["DZV5"] = pack(self.scales[2])  # DZ
        label["Z0V4"] = pack(self.offsets[2])  # Z0
        label["DYV3"] = pack(self.scales[1])  # DY
        label["Y0V2"] = pack(self.offsets[1])  # Y0
        label["DXV1"] = pack(self.scales[0])  # DX
        label["X0V0"] = pack(self.offsets[0])  # X0
        label["NTITLE"] = len(title)
        label["TITLE"][:, : len(title)] = [ord(s) for s in title]
        xunit = self.units[0] if self.units[0] is not None else ""
        label["XUNIT"][:, : len(xunit)] = [ord(c) for c in xunit]
        yunit = self.units[0] if self.units[0] is not None else ""
        label["YUNIT"][:, : len(yunit)] = [ord(c) for c in yunit]
        zunit = self.units[0] if self.units[0] is not None else ""
        label["ZUNIT"][:, : len(zunit)] = [ord(c) for c in zunit]
        return label

    @classmethod
    def _check_format(cls, data):
        if np.issubdtype(data.dtype, np.int8):
            iform = 0  # byte
        elif np.issubdtype(data.dtype, np.int16):
            iform = 1  # int16
        elif np.issubdtype(data.dtype, np.floating) and data.dtype.itemsize <= 4:
            data = data.astype(np.float32)
            iform = 2  # float (4 byte or less)
        elif np.issubdtype(data.dtype, np.complexfloating) and data.dtype.itemsize <= 8:
            data = data.astype(np.complex64)
            iform = 3  # complex (8 byte or less)
        elif np.issubdtype(data.dtype, np.int32):
            iform = 4  # int32
        else:
            supported_formats = [np.dtype(i).name for i in cls.IFORM_DICT.values()]
            msg = "The SEMPER file format does not support {} data type. ".format(
                data.dtype.name
            )
            msg += "Supported data types are: " + ", ".join(supported_formats)
            raise IOError(msg)
        return data, iform

    @classmethod
    def load_from_unf(cls, filename, lazy=False):
        r"""Load a `.unf`-file into a :class:`~.SemperFormat` object.

        Parameters
        ----------
        filename : string
            The name of the unf-file from which to load the data. Standard
            format is '\*.unf'.

        Returns
        -------
        semper : :class:`~.SemperFormat` (N=1)
            SEMPER file format object containing the loaded information.

        """
        metadata = OrderedDict()
        with open(filename, "rb") as f:
            # Read header:
            rec_length = np.fromfile(f, dtype="<i4", count=1)[0]  # length of header
            header = np.fromfile(f, dtype=cls.HEADER_DTYPES[: rec_length // 2], count=1)
            metadata.update(sarray2dict(header))
            assert np.frombuffer(f.read(4), dtype=np.int32)[0] == rec_length, (
                "Error while reading the header (length is not correct)!"
            )
            data_format = cls.IFORM_DICT[metadata["IFORM"]]
            iversn, remain = divmod(metadata["IFLAG"], 10000)
            ilabel, ntitle = divmod(remain, 1000)
            metadata.update({"IVERSN": iversn, "ILABEL": ilabel, "NTITLE": ntitle})
            # Read title:
            title = ""
            if ntitle > 0:
                assert (
                    np.fromfile(f, dtype="<i4", count=1)[0] == ntitle
                )  # length of title
                title = b"".join(np.fromfile(f, dtype="c", count=ntitle))
                title = title.decode()
                metadata["TITLE"] = title
                assert np.fromfile(f, dtype="<i4", count=1)[0] == ntitle
            if ilabel:
                try:
                    metadata.update(cls._read_label(f))
                except Exception as e:
                    warning = "Could not read label, trying to proceed without it!"
                    warning += " (Error message: {})".format(str(e))
                    warnings.warn(warning)
            # Read picture data:
            pos = f.tell()
            shape = metadata["NLAY"], metadata["NROW"], metadata["NCOL"]
            if lazy:
                from dask import delayed
                from dask.array import from_delayed

                task = delayed(_read_data)(f, filename, pos, data_format, shape)
                data = from_delayed(task, shape=shape, dtype=data_format)
            else:
                data = _read_data(f, filename, pos, data_format, shape)
        offsets = (
            metadata.get("X0V0", 0.0),
            metadata.get("Y0V2", 0.0),
            metadata.get("Z0V4", 0.0),
        )
        scales = (
            metadata.get("DXV1", 1.0),
            metadata.get("DYV3", 1.0),
            metadata.get("DZV5", 1.0),
        )
        units = (
            metadata.get("XUNIT", None),
            metadata.get("YUNIT", None),
            metadata.get("ZUNIT", None),
        )
        return cls(data, title, offsets, scales, units, metadata)

    def save_to_unf(self, filename, skip_header=False):
        """Save a :class:`~.SemperFormat` to a file.

        Parameters
        ----------
        filename : string
            The name of the unf-file to which the data should be written.
        skip_header : bool, optional
            Determines if the header, title and label should be skipped (useful
            for some other programs). Default is False.

        Returns
        -------
        None

        """
        nlay, nrow, ncol = self.data.shape
        data, iform = self._check_format(self.data)
        title = self.title if self.title is not None else ""
        with open(filename, "wb") as f:
            if not skip_header:
                # Create header:
                header = np.zeros(
                    (1,), dtype=self.HEADER_DTYPES[:-1]
                )  # IFORMAT is not used!
                # Fill header:
                header["NCOL"] = self.data.shape[2]
                header["NROW"] = self.data.shape[1]
                header["NLAY"] = self.data.shape[0]
                header["ICLASS"] = self.metadata.get("ICLASS", 6)  # 6: Undefined!
                header["IFORM"] = iform
                # IVERSN: 2; ILABEL:1 (True)
                header["IFLAG"] = 2 * 10000 + 1000 + len(title)
                # Write header:
                f.write(struct.pack("<i", 12))  # record length, 4 byte format!
                f.write(header.tobytes())
                f.write(struct.pack("<i", 12))  # record length, 4 byte format!
                # Write title:
                if len(title) > 0:
                    f.write(
                        struct.pack("<i", len(title))
                    )  # record length, 4 byte format!
                    f.write(title.encode())
                    f.write(
                        struct.pack("<i", len(title))
                    )  # record length, 4 byte format!
                # Create label:
                label = self._get_label()
                # Write label:
                f.write(struct.pack("<i", 2 * 256))  # record length, 4 byte format!
                f.write(label.tobytes())
                f.write(struct.pack("<i", 2 * 256))  # record length, 4 byte format!
            # Write picture data:
            for k in range(nlay):
                for j in range(nrow):
                    row = self.data[k, j, :]
                    # Record length = bytes per entry * ncol:
                    record_length = np.dtype(self.IFORM_DICT[iform]).itemsize * ncol
                    f.write(
                        struct.pack("<i", record_length)
                    )  # record length, 4 byte format!
                    f.write(row.tobytes())
                    # SEMPER always expects an even number of bytes per row,
                    # which is only a problem for writing single byte data
                    # (IFORM = 0, np.byte). If ncol is odd, an empty byte (0)
                    # is added:
                    if self.data.dtype == np.byte and ncol % 2 != 0:
                        np.zeros(1, dtype=np.byte).tobytes()
                    # record length, 4 byte format!
                    f.write(struct.pack("<i", record_length))

    @classmethod
    def from_signal(cls, signal):
        """Import a :class:`~.SemperFormat` object from a
        :class:`~hyperspy.signals.Signal` object.

        Parameters
        ----------
        signal: :class:`~hyperspy.signals.Signal`
            The signal which should be imported.

        Returns
        -------
        None

        """
        data = signal["data"]
        smetadata = DTBox(signal["metadata"], box_dots=True)
        assert len(data.shape) <= 3, "Only up to 3-dimensional datasets can be handled!"
        scales, offsets, units = [1.0] * 3, [0.0] * 3, [None] * 3  # Defaults!
        for i in range(len(data.shape)):
            scales[i] = signal["axes"][i]["scale"]
            offsets[i] = signal["axes"][i]["offset"]
            units[i] = signal["axes"][i]["units"]
        # Make sure data is 3D!
        data = data[tuple(None for _ in range(3 - len(data.shape)))]
        signal_dimension = len(
            [axis for axis in signal["axes"] if not axis["navigate"]]
        )
        if signal_dimension == 1:
            record_by = "spectrum"
        elif signal_dimension == 2:
            record_by = "image"
        else:
            record_by = ""
        iclass = cls.ICLASS_DICT_INV.get(record_by, 6)  # 6: undefined
        data, iform = cls._check_format(data)
        title = smetadata.General.to_dict().get("title", None)
        metadata = OrderedDict()
        if "date" in smetadata.General.keys() and "time" in smetadata.General.keys():
            dt = "%s %s" % (smetadata.General.date, smetadata.General.time)
        else:
            dt = "%s" % datetime.now()

        metadata.update(
            {
                "DATE": "%s" % dt.split(".")[0],
                "ICLASS": iclass,
                "IFORM": iform,
                "IVERSN": 2,  # Current standard
                "ILABEL": 1,  # True
                "IWP": 0,  # Seems standard
                "IPLTYP": 248,  # Seems standard
                "ICCOLN": data.shape[2] // 2 + 1,
                "ICROWN": data.shape[1] // 2 + 1,
                "ICLAYN": data.shape[0] // 2 + 1,
            }
        )
        return cls(data, title, offsets, scales, units, metadata)

    def to_dictionary(self, lazy=False):
        """Export a :class:`~.SemperFormat` object to a
        signal dictionary.

        Parameters
        ----------
        None

        Returns
        -------
        signal: dict
            The exported signal.

        """
        data = self.data.squeeze()  # Reduce unneeded dimensions!
        axes = []
        for name, scale, offset, units, size in zip(
            ("x", "y", "z"), self.scales, self.offsets, self.units, data.shape
        ):
            axes.append(
                {
                    "_type": "UniformDataAxis",
                    "name": name,
                    "units": units,
                    "navigate": False,
                    "is_binned": False,
                    "size": size,
                    "scale": scale,
                    "offset": offset,
                }
            )
        metadata = {}
        metadata["General"] = {}
        metadata["General"]["title"] = self.title
        if "DATE" in self.metadata.keys():
            date, time = self._convert_date_time_from_label()
            metadata["General"]["date"] = date
            metadata["General"]["time"] = time

        signal_dic = {
            "data": data,
            "metadata": metadata,
            "original_metadata": self.metadata,
            "axes": axes,
            "attributes": {"_lazy": lazy, "ragged": False},
        }
        return signal_dic

    def _convert_date_time_from_label(self):
        # Convert the label['DATE'] to ISO 8601 for metadata
        try:
            dt = datetime.strptime(self.metadata["DATE"], "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = datetime.strptime(self.metadata["DATE"], "%y-%m-%d %H:%M:%S")
        return dt.date().isoformat(), dt.time().isoformat()

    def log_info(self):
        """log important flag information of the :class:`.~SemperFormat`
        object.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        info_str = (
            "Semper info:\n"
            + ("\t%s\n" % self.title)
            + ("\tdimensions: x: {}, y: {}, z: {}\n".format(*reversed(self.data.shape)))
            + ("\tscaling:    x: {:.3g}, y: {:.3g}, z: {:.3g}\n".format(*self.scales))
            + ("\toffsets:    x: {:.3g}, y: {:.3g}, z: {:.3g}".format(*self.offsets))
            + ("\tunits:      x: {}, y: {}, z: {}".format(*self.units))
            + ("\tdata range: %s %s\n" % (self.data.min(), self.data.max()))
            + ("\tmetadata:\n")
        )
        for k, v in self.metadata.items():
            info_str += "\t\t{}: {}".format(k, v)
        _logger.info(info_str)


def unpack_from_intbytes(fmt, byte_list):
    """Read in a list of bytes (as int with range 0-255) and unpack them with
    format `fmt`.
    """
    return struct.unpack(fmt, b"".join(map(bytes, [[byte] for byte in byte_list])))[0]


def pack_to_intbytes(fmt, value):
    """Pack a `value` into a byte list using format `fmt` and represent it as
    int (range 0-255).
    """
    return [int(c) for c in struct.pack(fmt, value)]


def _read_data(fobj, fname, position, data_format, shape):
    if fobj.closed:
        fobj = open(fname, "rb")
    fobj.seek(position)
    nlay, nrow, ncol = shape
    data = np.empty(shape, dtype=data_format)
    for k in range(nlay):
        for j in range(nrow):
            rec_length = np.fromfile(fobj, dtype="<i4", count=1)[0]
            # Not always ncol, see below
            count = rec_length // np.dtype(data_format).itemsize
            row = np.fromfile(fobj, dtype=data_format, count=count)
            # [:ncol] is used because Semper always writes an even
            # number of bytes which is a problem when reading in single
            # bytes (IFORM = 0, np.byte). If ncol is odd, an empty
            # byte (0) is added which has to be skipped during read in:
            data[k, j, :] = row[:ncol]
            test = np.fromfile(fobj, dtype="<i4", count=1)[0]
            assert test == rec_length
    return data


def file_reader(filename, lazy=False):
    """
    Read a Semper ``.unf`` file.

    Parameters
    ----------
    %s
    %s

    %s
    """
    semper = SemperFormat.load_from_unf(filename, lazy=lazy)
    semper.log_info()
    return [semper.to_dictionary(lazy=lazy)]


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def file_writer(filename, signal, skip_header=False):
    """
    Write signal to a Semper ``.unf`` file.

    Parameters
    ----------
    %s
    %s
    skip_header : bool, default=False
        Determines if the header, title and label should be skipped (useful
        for some other programs).
    """
    semper = SemperFormat.from_signal(signal)
    semper.save_to_unf(filename, skip_header)


file_writer.__doc__ %= (FILENAME_DOC.replace("read", "write to"), SIGNAL_DOC)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from PIL.ExifTags import TAGS

CustomTAGS = {
    **TAGS,
    # Customized EXIF TAGS from Renishaw
    0xFEA0: "FocalPlaneXYOrigins",  # 65184
    0xFEA1: "FieldOfViewXY",  # 65185
    0xFEA2: "Unknown",  # 65186, could it be magnification?
}


# from https://exiftool.org/TagNames/EXIF.html
# For tag 0x9210 (37392)
FocalPlaneResolutionUnit_mapping = {
    None: None,
    1: None,
    2: "inches",
    3: "cm",
    4: "mm",
    5: "µm",
}


def _parse_axes_from_metadata(exif_tags, sizes):
    # return of axes must not be empty, or dimensions are lost
    # if no exif_tags exist, axes are set to a scale of 1 per pixel,
    # unit is set to None, hyperspy will parse it as a traits.api.undefined value
    offsets = [0, 0]
    fields_of_views = [sizes[1], sizes[0]]
    unit = None
    if exif_tags is not None:
        # Fallback to default value when tag not available
        offsets = exif_tags.get("FocalPlaneXYOrigins", offsets)
        # jpg files made with Renishaw have this tag
        fields_of_views[0] = exif_tags.get("FocalPlaneXResolution", fields_of_views[0])
        fields_of_views[1] = exif_tags.get("FocalPlaneYResolution", fields_of_views[1])
        unit = FocalPlaneResolutionUnit_mapping[
            exif_tags.get("FocalPlaneResolutionUnit", unit)
        ]

    axes = [
        {
            "name": name,
            "units": unit,
            "size": size,
            "scale": fields_of_views[i] / size,
            "offset": offsets[i],
            "index_in_array": i,
        }
        for i, name, size in zip([1, 0], ["y", "x"], sizes)
    ]

    return axes


def _parse_exif_tags(im):
    """
    Parse exif tags from a pillow image

    Parameters
    ----------
    im : :class:`PIL.Image`
        The pillow image from which the exif tags will be parsed.

    Returns
    -------
    exif_dict : None or dict
        The dictionary of exif tags.

    """
    exif_dict = None
    try:
        # missing header keys when Pillow >= 8.2.0 -> does not flatten IFD anymore
        # see https://pillow.readthedocs.io/en/stable/releasenotes/8.2.0.html#image-getexif-exif-and-gps-ifd
        # Use fall-back _getexif method instead
        # Not all format plugin have the private method
        # prefer to use that method as it returns more items
        exif_dict = im._getexif()
    except AttributeError:
        exif_dict = im.getexif()
    if exif_dict is not None:
        exif_dict = {CustomTAGS.get(k, "unknown"): v for k, v in exif_dict.items()}

    return exif_dict
